
from datetime import datetime

import typeguard

from cloud_governance.common.clouds.aws.athena.pyathena_operations import PyAthenaOperations
from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.main.environment_variables import environment_variables


class SpotSavingsAnalysis:
    """
    This class contain the spt savings analysis reports from the athena query
    that are gathered from the AWS Cost and Usage Reports.
    To get reports from the athena:
    1. Enable the cost-and-usage reports with support of athena integration
    2. Create a Database and table of CUR
    """

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__default_round_digits = self.__environment_variables_dict.get('DEFAULT_ROUND_DIGITS')
        self.__es_index = self.__environment_variables_dict.get('es_index')
        self.__database_name = self.__environment_variables_dict.get('ATHENA_DATABASE_NAME')
        self.__table_name = self.__environment_variables_dict.get('ATHENA_TABLE_NAME')
        self.__es_operations = ElasticSearchOperations()

    def __get_prepared_query(self):
        """
        This method prepare the query
        :return:
        """
        current_date = datetime.utcnow()
        year = current_date.year
        current_month = current_date.month
        previous_month = current_month - 1 if current_month - 1 != 0 else 12
        query = f"""
                SELECT
                date_format(line_item_usage_start_date, '%Y-%m-%d') as CurrentDate,
                date_format(bill_billing_period_start_date, '%Y-%m-%d') as  MonthStartDate,
                line_item_usage_account_id as AccountId,
                line_item_product_code as ProductCode,
                product_region as Region,
                product_instance_type as InstanceType,
                cost_category_cost_center as CostCenter,
                cost_category_o_us as CostCategory,
                cost_category_organization as RHOrg,
                ROUND(SUM(discount_total_discount), 3) as TotalDiscount,
                ROUND(SUM(line_item_usage_amount), {self.__default_round_digits}) as UsageAmount,
                ROUND(SUM(line_item_unblended_cost + discount_total_discount), {self.__default_round_digits}) as UnblendedCost,
                ROUND(SUM(pricing_public_on_demand_cost), {self.__default_round_digits}) as OnDemand,
                ROUND(SUM(pricing_public_on_demand_cost - line_item_unblended_cost), {self.__default_round_digits}) as SpotSavings
                FROM "{self.__database_name}"."{self.__table_name}"
                WHERE "product_product_name" = 'Amazon Elastic Compute Cloud'
                AND "line_item_resource_id" LIKE 'i-%'
                AND "line_item_operation" LIKE 'RunInstance%'
                AND "product_marketoption" = 'Spot'
                AND month(bill_billing_period_start_date) in ({previous_month}, {current_month})
                AND year(bill_billing_period_start_date) = {year}
                AND pricing_public_on_demand_cost <> 0
                GROUP BY 1, 2, 3, 4, 5, 6, 7, 8, 9 ORDER BY MonthStartDate desc
                """
        return query

    def __get_query_results(self):
        """
        This method get queries data
        :return:
        """
        result = {}
        query_string = self.__get_prepared_query()
        if query_string:
            athena_operations = PyAthenaOperations()
            result = athena_operations.execute_query(query_string)
        else:
            logger.debug(f"query string is not provided, exit without query execution")
        return result

    @typeguard.typechecked
    def __get_data_to_upload_to_es(self, athena_data_dictionary: list):
        """
        This method returns ready upload dict to upload to elasticsearch
        :return:
        """
        for row_dict in athena_data_dictionary:
            month_start_date = row_dict.get('MonthStartDate')
            month_name_year = datetime.strftime(datetime.strptime(month_start_date, '%Y-%m-%d'), '%Y %b')
            row_dict['Month'] = month_name_year
            row_dict['filter_date'] = f'{month_start_date}-{month_name_year.split()[-1]}'
            row_dict['AccountIdCostCenter'] = f'{row_dict.get("AccountId")}-{row_dict.get("CostCenter")}'
            row_dict['index_id'] = f'{row_dict.get("CurrentDate")}-' \
                                   f'{row_dict.get("AccountId")}-' \
                                   f'{row_dict.get("Region")}-{row_dict.get("InstanceType")}'
            row_dict['AWSCostCenter'] = f'AWS-{row_dict.get("CostCenter")}'

    @logger_time_stamp
    def __collect_reports_and_upload_es(self):
        """
        This method collects the data and uploads to elastic search
        :return:
        """
        query_result = self.__get_query_results()
        if query_result:
            self.__get_data_to_upload_to_es(athena_data_dictionary=query_result)
            if query_result:
                self.__es_operations.upload_data_in_bulk(data_items=query_result, id='index_id', index=self.__es_index)

    @logger_time_stamp
    def run(self):
        """
        This is the starting of the methods
        :return:
        """
        self.__collect_reports_and_upload_es()



import datetime
import re

from math import ceil

from cloud_governance.cloud_resource_orchestration.utils.common_operations import string_equal_ignore_case
from cloud_governance.common.utils.configs import GRAVITON_MAPPINGS, DEFAULT_ROUND_DIGITS, DEFAULT_GRAVITON_INSTANCE
from cloud_governance.policy.helpers.aws.aws_policy_operations import AWSPolicyOperations


class ClusterRun(AWSPolicyOperations):
    """
    This class performs the operations on running cluster resources
    """

    def __init__(self):
        super().__init__()

    def __get_creation_date(self, block_devices: list) -> str:
        """
        This method returns the creation date of the EC2 instance from the root ebs volume.
        :param block_devices:
        :return:
        """
        for devices in block_devices:
            if devices.get('Ebs', {}).get('DeleteOnTermination'):
                return devices.get('Ebs', {}).get('AttachTime').strftime('%Y-%m-%d')
        return ''

    def run_policy_operations(self):
        """
        This method returns the running vms in the AAzure cloud and stops based on the action
        :return:
        :rtype:
        """
        instances = self._get_all_instances()
        cluster_data = {}
        for instance in instances:
            tags = instance.get('Tags', [])
            cluster_tag = self._get_cluster_tag(tags=tags).strip()
            instance_state = instance.get('State', {}).get('Name')
            if cluster_tag and not string_equal_ignore_case(instance_state, 'terminated'):
                name_tag = self.get_tag_name_from_tags(tags=tags, tag_name='Name')
                launch_time = instance.get('LaunchTime')
                running_instances = stopped_instances = 0
                running_days = self.calculate_days(launch_time)
                running_hours = ceil(self.calculate_hours(launch_time))
                stopped_date_time = ''
                if string_equal_ignore_case(instance_state, 'stopped'):
                    stopped_instances = 1
                    state_transition_reason = instance.get('StateTransitionReason')
                    if state_transition_reason:
                        extract_data = re.search(r'\((\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})', state_transition_reason)
                        if extract_data:
                            stopped_date_time = extract_data.group(1)
                            running_days = self.calculate_days(stopped_date_time.split()[0], start_date=launch_time)
                            instance_state += f"@{extract_data.group(1)}"
                else:
                    running_instances = 1
                creation_date = ''
                rosa_cluster = True if self.get_tag_name_from_tags(tags=tags, tag_name='red-hat-managed') else False
                instance_type = instance.get('InstanceType')
                using_graviton = False
                for graviton_instance_family in GRAVITON_MAPPINGS.values():
                    if graviton_instance_family in instance_type:
                        using_graviton = True
                if 'master' in name_tag.lower():
                    creation_date = self.__get_creation_date(instance.get('BlockDeviceMappings', []))
                instance_data = f"{instance.get('InstanceId')}, {self.get_tag_name_from_tags(tags=tags, tag_name='Name')}, {instance.get('InstanceType')}, {instance_state}, {running_days}, {launch_time}"
                if cluster_tag in cluster_data:
                    if creation_date:
                        cluster_data[cluster_tag]['CreationDate'] = creation_date
                        cluster_data[cluster_tag]['ClusterState'] = instance_state
                        cluster_data[cluster_tag]['StoppedDate'] = stopped_date_time
                    cluster_data[cluster_tag]['Instances'].append(instance_data)
                    cluster_data[cluster_tag]['InstanceTypes'].append(instance.get('InstanceType'))
                    cluster_data[cluster_tag]['InstanceCount'] = len(cluster_data[cluster_tag]['Instances'])
                    cluster_data[cluster_tag]['Stopped'] = int(cluster_data[cluster_tag]['Stopped']) + stopped_instances
                    cluster_data[cluster_tag]['Running'] = int(cluster_data[cluster_tag]['Running']) + running_instances
                else:
                    cluster_data[cluster_tag] = {
                        'ClusterName': cluster_tag.split('/')[-1].upper(),
                        'ClusterName2': cluster_tag.split('/')[-1].lower(),
                        'ResourceId': cluster_tag,
                        'ClusterTag': cluster_tag,
                        'InstanceTypes': [instance.get('InstanceType')],
                        'User': self.get_tag_name_from_tags(tags=tags, tag_name='User'),
                        'RunningDays': running_days,
                        'RunningHours': running_hours,
                        'RegionName': self._region,
                        'PublicCloud': self._cloud_name,
                        'Instances': [instance_data],
                        'LaunchTime': launch_time.date(),
                        'InstanceCount': 1,
                        'Stopped': stopped_instances,
                        'Running': running_instances,
                        'Graviton': using_graviton,
                        'RosaCluster': rosa_cluster,
                        'index-id': f'{datetime.datetime.now(datetime.timezone.utc).date()}-{self._cloud_name.lower()}-{self.account.lower()}-{self._region.lower()}-{cluster_tag}',
                    }
                    if creation_date:
                        cluster_data[cluster_tag]['creation_date'] = creation_date
                        cluster_data[cluster_tag]['ClusterState'] = instance_state
                        cluster_data[cluster_tag]['StoppedDate'] = stopped_date_time
        for cluster in cluster_data.values():
            instance_types = cluster['InstanceTypes']
            total_cost = 0
            graviton_instance_cost = 0
            cluster['GravitonInstanceTypes'] = []
            running_hours = 1 if cluster['RunningHours'] == 0 else cluster['RunningHours']
            for instance_type in set(instance_types):
                instance_types_count = instance_types.count(instance_type)
                unit_price = self._resource_pricing.get_ec2_price(region_name=self._region,
                                                                  instance_type=instance_type)
                total_cost += (unit_price * running_hours * instance_types_count)
                if not cluster['Graviton']:
                    instance_family, instance_size = instance_type.split('.')
                    graviton_instance = f'{DEFAULT_GRAVITON_INSTANCE}.{instance_size}'
                    if instance_family in GRAVITON_MAPPINGS:
                        graviton_instance = f'{GRAVITON_MAPPINGS[instance_family]}.{instance_size}'
                    graviton_unit_price = self._resource_pricing.get_ec2_price(region_name=self._region,
                                                                               instance_type=graviton_instance)
                    graviton_instance_cost += (graviton_unit_price * running_hours * instance_types_count)
                    cluster['GravitonInstanceTypes'].append(f"{graviton_instance}: {instance_types_count}")

            cluster['TotalGravitonInstanceCost'] = round(graviton_instance_cost, DEFAULT_ROUND_DIGITS)
            cluster['TotalCost'] = round(total_cost, DEFAULT_ROUND_DIGITS)

            cluster['GravitonSavings'] = round(total_cost - graviton_instance_cost,
                                               DEFAULT_ROUND_DIGITS) if graviton_instance_cost != 0 else 0
            cluster['InstanceTypes'] = [f"{x}: {instance_types.count(x)}" for x in set(instance_types)]
        return list(cluster_data.values())



from ibm_platform_services.global_tagging_v1 import GlobalTaggingV1, Resource

from cloud_governance.common.clouds.ibm.account.ibm_authenticator import IBMAuthenticator
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp


class GlobalTaggingOperations(IBMAuthenticator):
    """
    This class performs tagging operations on cloud resources.
    """
    BATCH_SIZE = 100

    def __init__(self):
        super().__init__()
        self.__tag_service = GlobalTaggingV1(authenticator=self.iam_authenticator)

    @logger_time_stamp
    def update_tags(self, resources_crn: list, tags: list):
        """
        This method updates the tags associated with an instance.
        :param resources_crn:
        :param tags:
        :return:
        """
        resources_list = [Resource(resource_crn) for resource_crn in resources_crn]
        resources_batch_list = [resources_list[i:i + self.BATCH_SIZE]
                                for i in range(0, len(resources_list), self.BATCH_SIZE)]
        success = 0
        errors = []
        tag_names = []
        for tag in tags:
            key, value = tag.split(":")
            tag_names.append(f'{key.strip()}:{value.strip()}')
        logger.info(f"Tagging {len(resources_crn)} resources.")
        for resource_batch in resources_batch_list:
            responses = self.__tag_service.attach_tag(resources=resource_batch, tag_names=tag_names) \
                .get_result()['results']
            for resource in responses:
                if resource['is_error']:
                    errors.append(resource.get('resource_id'))
                    logger.error(f'Unable to attach resource tags to: {resource["resource_id"]}')
                else:
                    success += 1
        return success == len(resources_crn), errors



"""
Policy to delete inactive IAM access keys older than DELETE_ACCESS_KEY_DAYS.
"""
from cloud_governance.common.utils.configs import DELETE_ACCESS_KEY_DAYS
from cloud_governance.policy.helpers.aws.aws_policy_operations import AWSPolicyOperations


class DeleteAccessKey(AWSPolicyOperations):
    RESOURCE_ACTION = "Delete"

    def run_policy_operations(self):
        """
        For inactive keys with age > DELETE_ACCESS_KEY_DAYS: apply a grace period
        (deletion_grace_days = age_days - DELETE_ACCESS_KEY_DAYS, capped at DAYS_TO_TAKE_ACTION). During
        grace period write to ES with cleanup_days 1..7 so send_aggregated_alerts sends
        reminder emails. After grace period, delete the key.
        """
        result = []
        days_to_take_action = int(self._days_to_take_action)
        iam_users_access_keys = self._get_iam_users_access_keys()

        for username, user_data in iam_users_access_keys.items():
            tags = user_data.get('tags', user_data.get('Tags', []))
            region = user_data['region']
            user_name = username

            for access_key_label, access_key_data in user_data.items():
                if 'access key' not in access_key_label.lower():
                    continue
                age_days = access_key_data.get('age_days')
                status = (access_key_data.get('status') or '').lower()
                if age_days is None or status != 'inactive':
                    continue
                age_days = int(age_days)

                key_num = access_key_label.split()[-1]
                inactive_tag_key = f"UnusedAccessKey{key_num}InactiveDate"
                inactive_date_str = self.get_tag_name_from_tags(tags=tags, tag_name=inactive_tag_key)
                delete_all_inactive = self._environment_variables_dict.get('DELETE_INACTIVE_KEYS_WITHOUT_TAG', False)
                if not (age_days > DELETE_ACCESS_KEY_DAYS and (inactive_date_str or delete_all_inactive)):
                    continue

                deletion_grace_days = min(age_days - DELETE_ACCESS_KEY_DAYS, days_to_take_action)
                cleanup_result = self.verify_and_delete_resource(
                    resource_id=user_name,
                    tags=tags,
                    clean_up_days=deletion_grace_days,
                    access_key_label=access_key_label,
                    access_key_id=access_key_data.get('access_key_id'),
                    remove_inactive_tag=bool(inactive_date_str),
                )
                resource_data = self._get_es_schema(
                    resource_id=user_name,
                    user=self.get_tag_name_from_tags(tags=tags, tag_name='User'),
                    skip_policy=self.get_skip_policy_value(tags=tags),
                    cleanup_days=deletion_grace_days,
                    dry_run=self._dry_run,
                    name=user_name,
                    region=region,
                    cleanup_result=str(cleanup_result),
                    resource_action=self.RESOURCE_ACTION,
                    cloud_name=self._cloud_name,
                    resource_type='UnusedAccessKey',
                    resource_state='Inactive',
                    age_days=age_days,
                    last_activity_days=access_key_data.get('last_activity_days'),
                    unit_price=0,
                )
                result.append(resource_data)

        return result




import asyncio
import logging
from urllib.parse import quote

import aiohttp
import urllib3
from aiohttp import BasicAuth

urllib3.disable_warnings()

logger = logging.getLogger(__name__)


class JiraException(Exception):
    pass


class Jira(object):
    def __init__(
        self,
        url,
        username,
        ticket_queue,
        password=None,
        token=None,
        loop=None,
    ):
        try:
            logger.debug(":Initializing Jira object:")
            self.url = url
            self.username = username
            self.ticket_queue = ticket_queue
            self.password = password
            self.token = token
            self._auth = None
            if not loop:
                self.loop = asyncio.new_event_loop()
                self.new_loop = True
            else:
                self.loop = loop
                self.new_loop = False

            # Jira Cloud API tokens use HTTP Basic auth (email + token), same as curl -u user:token
            if self.token:
                if not self.username:
                    logger.error(
                        "Basic Authentication with API token requires JIRA_USERNAME (email)"
                    )
                    raise JiraException(
                        "Basic Authentication with API token requires JIRA_USERNAME (email)"
                    )
                self._auth = BasicAuth(self.username, self.token)
            elif self.password:
                if not self.username:
                    logger.error(
                        "Basic Authentication requires JIRA_USERNAME"
                    )
                    raise JiraException(
                        "Basic Authentication requires JIRA_USERNAME"
                    )
                self._auth = BasicAuth(self.username, self.password)
            else:
                logger.error(
                    "Basic Authentication expected: set JIRA_TOKEN or JIRA_PASSWORD"
                )
                raise JiraException(
                    "Basic Authentication expected: set JIRA_TOKEN or JIRA_PASSWORD"
                )
            self.headers = {"Accept": "application/json"}
        except JiraException:
            raise
        except Exception as ex:
            logger.error("Failed to initialize Jira client: %s", ex)
            raise JiraException from ex

    def __exit__(self):
        if self.new_loop:
            self.loop.close()

    async def get_request(self, endpoint):
        logger.debug("GET: %s" % endpoint)
        try:
            async with aiohttp.ClientSession(
                auth=self._auth,
                headers=self.headers,
                loop=self.loop,
            ) as session:
                async with session.get(
                    self.url + endpoint,
                    verify_ssl=False,
                ) as response:
                    result = await response.json(content_type="application/json")
        except Exception as ex:
            logger.debug(ex)
            logger.error("There was something wrong with your request.")
            return None
        if response.status == 404:
            logger.error("Resource not found: %s" % self.url + endpoint)
            return None
        return result

    async def post_request(self, endpoint, payload):
        logger.debug("POST: {%s:%s}" % (endpoint, payload))
        try:
            async with aiohttp.ClientSession(
                auth=self._auth,
                headers=self.headers,
                loop=self.loop,
            ) as session:
                async with session.post(
                    self.url + endpoint,
                    json=payload,
                    verify_ssl=False,
                ) as response:
                    if response.status == 204:
                        data = {}
                    else:
                        data = await response.json(content_type="application/json")
        except Exception as ex:
            logger.debug(ex)
            logger.error("There was something wrong with your request.")
            return False
        if response.status in [200, 201, 204]:
            logger.info("Post successful.")
            return data
        if response.status == 404:
            logger.error("Resource not found: %s" % self.url + endpoint)
        else:
            logger.error(data)
        return False

    async def put_request(self, endpoint, payload):
        logger.debug("POST: {%s:%s}" % (endpoint, payload))
        try:
            async with aiohttp.ClientSession(
                auth=self._auth,
                headers=self.headers,
                loop=self.loop,
            ) as session:
                async with session.put(
                    self.url + endpoint,
                    json=payload,
                    verify_ssl=False,
                ) as response:
                    await response.json(content_type="application/json")
        except Exception as ex:
            logger.debug(ex)
            logger.error("There was something wrong with your request.")
            return False
        if response.status in [200, 201, 204]:
            logger.info("Post successful.")
            return True
        if response.status == 404:
            logger.error("Resource not found: %s" % self.url + endpoint)
        return False

    async def create_ticket(self, summary, description):
        endpoint = "/issue/"
        logger.debug("POST new ticket")
        short_summary = summary.split('\r')
        title = f"{short_summary[0]}"

        data = {
            "fields": {
                "project": {"key": self.ticket_queue},
                "issuetype": {"name": "Task"},
                "summary": title,
                "description": description,
            }
        }
        response = await self.post_request(endpoint, data)
        return response

    async def create_subtask(self, parent_ticket, cloud, description, type_of_subtask):
        endpoint = "/issue/"
        logger.debug("POST new subtask")
        title = f"{cloud} {type_of_subtask}"

        data = {
            "fields": {
                "project": {"key": self.ticket_queue},
                "issuetype": {"id": "5"},
                "parent": {"key": f"{self.ticket_queue}-{parent_ticket}"},
                "summary": title,
                "labels": [type_of_subtask.upper()],
                "description": description,
            }
        }
        response = await self.post_request(endpoint, data)
        return response

    async def _resolve_watcher_to_account_id(self, watcher):
        """
        Jira Cloud expects an Atlassian accountId in the watchers POST body (JSON string),
        not an email. Resolve email or short username via REST v2 user search.
        """
        if not watcher or not str(watcher).strip():
            return None
        w = str(watcher).strip()
        user = await self.get_user_by_email(w)
        if user and user.get("accountId"):
            return user["accountId"]
        if "@" not in w:
            user = await self.get_user_by_email(f"{w}@redhat.com")
            if user and user.get("accountId"):
                return user["accountId"]
        logger.error("Could not resolve Jira accountId for watcher: %s", w)
        return None

    async def add_watcher(self, ticket, watcher):
        account_id = await self._resolve_watcher_to_account_id(watcher)
        if not account_id:
            return False
        issue_id = "%s-%s" % (self.ticket_queue, ticket)
        endpoint = "/issue/%s/watchers" % issue_id
        logger.debug(
            "POST watcher issue=%s accountId=%s (from %s)",
            issue_id,
            account_id,
            watcher,
        )
        response = await self.post_request(endpoint, account_id)
        return response

    async def add_label(self, ticket, label):
        issue_id = "%s-%s" % (self.ticket_queue, ticket)
        endpoint = "/issue/%s" % issue_id
        data = {"update": {"labels": [{"add": label}]}}
        response = await self.put_request(endpoint, data)
        return response

    async def post_comment(self, ticket, comment):
        issue_id = "%s-%s" % (self.ticket_queue, ticket)
        endpoint = "/issue/%s/comment" % issue_id
        payload = {"body": comment}
        response = await self.post_request(endpoint, payload)
        return response

    async def post_transition(self, ticket, transition):
        issue_id = "%s-%s" % (self.ticket_queue, ticket)
        endpoint = "/issue/%s/transitions" % issue_id
        payload = {"transition": {"id": transition}}
        response = await self.post_request(endpoint, payload)
        return response

    async def get_transitions(self, ticket):
        issue_id = "%s-%s" % (self.ticket_queue, ticket)
        endpoint = "/issue/%s/transitions" % issue_id
        result = await self.get_request(endpoint)
        if not result:
            logger.error("Failed to get transitions")
            return []

        transitions = result.get("transitions")
        if transitions:
            return transitions
        else:
            logger.error("No transitions found under %s" % issue_id)
            return []

    async def get_ticket(self, ticket):
        issue_id = "%s-%s" % (self.ticket_queue, ticket)
        endpoint = "/issue/%s" % issue_id
        result = await self.get_request(endpoint)
        if not result:
            logger.error("Failed to get ticket")
            return None
        return result

    async def get_watchers(self, ticket):
        issue_id = "%s-%s" % (self.ticket_queue, ticket)
        endpoint = "/issue/%s/watchers" % issue_id
        logger.debug("GET watchers: %s" % endpoint)
        result = await self.get_request(endpoint)
        if not result:
            logger.error("Failed to get watchers")
            return None
        return result

    async def get_user_by_email(self, email):
        endpoint = f"/user/search?query={quote(email)}"
        logger.debug("GET user: %s" % endpoint)
        result = await self.get_request(endpoint)
        if not result:
            logger.error("User not found for query: %s", email)
            return None
        el = email.lower()
        for user in result:
            if (user.get("emailAddress") or "").lower() == el:
                return user
        if len(result) == 1:
            return result[0]
        return None

    async def get_pending_tickets(self):
        query = {"statusCategory": 4, "labels": "EXTENSION"}
        logger.debug("GET pending tickets")
        result = await self.search_tickets(query)
        if not result:
            logger.error("Failed to get pending tickets")
            return None
        return result

    async def search_tickets(self, query=None):
        project = {"project": self.ticket_queue}
        prefix = "/search?jql="
        query_items = []

        if not query:
            query = project
        else:
            query.update(project)

        for k, v in query.items():
            query_items.append(f"{k}={v}")

        jql = " AND ".join(query_items)
        endpoint = f"{prefix}{jql}"
        logger.debug("GET pending tickets: %s" % endpoint)
        result = await self.get_request(endpoint)
        if not result:
            logger.error("Failed to get pending tickets")
            return None
        return result



from azure.mgmt.resource import ResourceManagementClient
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.clouds.azure.subscriptions.azure_operations import AzureOperations


class TagAzureResourceGroup:
    def __init__(self):
        """
        Initialize Azure clients and set the tags to apply.
        """
        self.__azure_operations = AzureOperations()
        self.__subscription_id = self.__azure_operations.subscription_id
        self.__credential = self.__azure_operations._AzureOperations__default_creds  # reuse credentials
        self.resource_client = ResourceManagementClient(self.__credential, self.__subscription_id)
        self.__tags_to_add = self.__azure_operations.global_tags
        if not self.__tags_to_add:
            raise ValueError("No tags provided to add. Please set GLOBAL_TAGS environment variable.")

    def tag_all(self):
        """
        Tag all resource groups and their contained resources with the specified tags.
        """
        resource_groups = self.resource_client.resource_groups.list()

        for rg in resource_groups:
            logger.info(f"Processing resource group: {rg.name}")
            # Tag the resource group
            try:
                self.resource_client.resource_groups.update(
                    resource_group_name=rg.name,
                    parameters={"tags": self.__tags_to_add}
                )
                logger.info(f"Tagged resource group: {rg.name}")
            except Exception as e:
                logger.error(f"Failed to tag resource group {rg.name}: {e}")
                continue

            # Tag all resources in the resource group
            for resource in self.resource_client.resources.list_by_resource_group(rg.name):
                logger.info(f"Tagging resource: {resource.name} ({resource.type})")

                existing_tags = resource.tags or {}
                updated_tags = {**existing_tags, **self.__tags_to_add}

                try:
                    parts = resource.type.split('/')
                    provider_ns = parts[0]
                    resource_type_str = '/'.join(parts[1:])

                    provider = self.resource_client.providers.get(provider_ns)

                    # Some resource_type may not match exactly, use a relaxed matching
                    resource_type_info = next(
                        (rt for rt in provider.resource_types if rt.resource_type.lower() == resource_type_str.lower()),
                        None
                    )

                    if not resource_type_info:
                        logger.warning(f"Could not find exact match for resource type: {resource_type_str} in provider {provider_ns}")
                        continue

                    api_version = next(
                        (v for v in resource_type_info.api_versions if 'preview' not in v.lower()),
                        resource_type_info.api_versions[0]
                    )

                    poller = self.resource_client.resources.begin_update_by_id(
                        resource_id=resource.id,
                        api_version=api_version,
                        parameters={
                            'location': resource.location,
                            'tags': updated_tags
                        }
                    )
                    poller.result()
                    logger.info(f"Tagged resource: {resource.name}")
                except Exception as e:
                    logger.error(f"Failed to tag resource {resource.name}: {e}")

    def run(self):
        """
        This method tags all Azure resources.
        """
        self.tag_all()



import typeguard

from cloud_governance.cloud_resource_orchestration.clouds.common.abstract_tagging_operations import \
    AbstractTaggingOperations
from cloud_governance.common.clouds.aws.resource_tagging_api.resource_tag_api_operations import ResourceTagAPIOperations
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp


class AWSTaggingOperations(AbstractTaggingOperations):
    """
    This class is performing the tagging operations on AWS
    """

    def __init__(self, region_name: str):
        super(AbstractTaggingOperations).__init__()
        self.__resource_tag_api_operations = ResourceTagAPIOperations(region_name=region_name)

    @logger_time_stamp
    def tag_resources_list(self, resources_list: list, update_tags_dict: dict):
        """
        This method updates the tags to the resources
        :param resources_list:
        :param update_tags_dict:
        :return:
        """
        self.__resource_tag_api_operations.tag_resources(resource_arn_list=resources_list,
                                                         update_tags_dict=update_tags_dict)

    def untag_resources_list(self, resources_list: list, delete_tag_keys_list: list):
        """
        This method untags the resources
        :param resources_list:
        :param delete_tag_keys_list:
        :return:
        """
        self.__resource_tag_api_operations.untag_resources(resource_arn_list=resources_list,
                                                           delete_tags_keys=delete_tag_keys_list)

    @typeguard.typechecked
    @logger_time_stamp
    def get_resources_list(self, tag_name: str, tag_value: str = ''):
        """
        This method returns all the resources having the tag_name and tag_value
        :param tag_name:
        :param tag_value:
        :return:
        """
        resources_list = self.__resource_tag_api_operations.get_resources(tag_name=tag_name, tag_value=tag_value)
        if resources_list:
            return resources_list
        return []



from cloud_governance.common.utils.utils import Utils
from cloud_governance.main.environment_variables import environment_variables
from cloud_governance.policy.policy_runners.azure.policy_runner import PolicyRunner as AzurePolicyRunner
from cloud_governance.policy.policy_runners.aws.policy_runner import PolicyRunner as AWSPolicyRunner
from cloud_governance.policy.policy_runners.ibm.policy_runner import PolicyRunner as IBMPolicyRunner


class MainOperations:

    def __init__(self):
        self.utils = Utils()
        self._environment_variables_dict = environment_variables.environment_variables_dict
        self._policy = self._environment_variables_dict.get('policy', '')
        self._public_cloud_name = self._environment_variables_dict.get('PUBLIC_CLOUD_NAME', '')

    def get_policy_runner(self):
        """
        This method returns the cloud policy runner object
        :return:
        :rtype:
        """
        policy_runner = None
        if Utils.equal_ignore_case(self._public_cloud_name, 'AWS'):
            policy_runner = AWSPolicyRunner()
        elif Utils.equal_ignore_case(self._public_cloud_name, 'Azure'):
            policy_runner = AzurePolicyRunner()
        else:
            if Utils.equal_ignore_case(self._public_cloud_name, 'IBM'):
                policy_runner = IBMPolicyRunner()

        return policy_runner

    def run(self):
        """
        This method run the AWS Policy operations
        :return:
        :rtype:
        """
        policies_list = Utils.get_cloud_policies(cloud_name=self._public_cloud_name, dir_dict=True)
        policy_runner = self.get_policy_runner()
        for policy_type, policies in policies_list.items():
            # @Todo support for all the aws policies, currently supports ec2_run as urgent requirement
            if self._policy in policies and self._policy in ["instance_run", "unattached_volume", "cluster_run",
                                                             "ip_unattached", "unused_nat_gateway", "instance_idle",
                                                             "zombie_snapshots", "database_idle", "s3_inactive", "unused_access_key",
                                                             "delete_access_key", "empty_roles", "tag_resources", "cost_usage_reports"]:
                source = policy_type
                if Utils.equal_ignore_case(policy_type, self._public_cloud_name):
                    source = ''
                policy_runner.run(source=source)
                return True
        return False



import os
import sys
import logging

logger_category_name = 'cloud_governance'
logger = logging.getLogger(logger_category_name)  # instantiating a logger
handler = logging.StreamHandler(sys.stdout)

account_name = os.environ.get('account')
log_format = f'[%(levelname)s] %(asctime)s {account_name} - %(message)s'
formatter = logging.Formatter(log_format)
handler.setFormatter(formatter)
logger.addHandler(handler)

# log for output only
log_path = '/tmp'
if os.environ.get('LOG_FILE_PATH'):
    log_path = os.environ.get('LOG_FILE_PATH')
fileHandler = logging.FileHandler(filename=f'{log_path}/cloud_governance.log', mode='w+')
fileHandler.setFormatter(fmt=formatter)
# logger.addHandler(fileHandler)

# def get_pyperf_log_path():
#     return f'{log_path}/py_perf.log'



from abc import ABC, abstractmethod

from cloud_governance.common.clouds.azure.compute.compute_operations import ComputeOperations
from cloud_governance.common.clouds.azure.compute.resource_group_operations import ResourceGroupOperations


class AbstractResource(ABC):

    def __init__(self):
        self._resource_group_operations = ResourceGroupOperations()
        self._compute_client = ComputeOperations()

    @abstractmethod
    def run(self):
        pass



from cloud_governance.common.utils.configs import UNUSED_ACCESS_KEY_DAYS
from cloud_governance.policy.helpers.aws.aws_policy_operations import AWSPolicyOperations


class UnusedAccessKey(AWSPolicyOperations):
    RESOURCE_ACTION = "DeActivate"

    def run_policy_operations(self):
        """
        For key age >= UNUSED_ACCESS_KEY_DAYS (e.g. 90): apply a grace period (deactivation_grace_days
        = age_days - UNUSED_ACCESS_KEY_DAYS, capped at DAYS_TO_TAKE_ACTION). During grace period
        write to ES with cleanup_days 1..7 so send_aggregated_alerts sends reminder emails. After
        grace period, deactivate the key.
        """
        unused_access_keys = []
        days_to_take_action = int(self._days_to_take_action)
        iam_users_access_keys = self._get_iam_users_access_keys()
        for username, user_data in iam_users_access_keys.items():
            tags = user_data.get('tags', user_data.get('Tags', []))
            region = user_data['region']
            user_name = username

            for access_key_label, access_key_data in user_data.items():
                if 'access key' not in access_key_label.lower():
                    continue
                last_activity_days = access_key_data.get('last_activity_days')
                age_days = access_key_data.get('age_days')
                status = (access_key_data.get('status') or '').lower()
                if age_days is None:
                    continue
                age_days = int(age_days)

                if status == 'inactive':
                    continue

                if not self._has_active_access_keys(user_name, access_key_label):
                    continue
                if self.get_skip_policy_value(tags=tags) in ('NOTDELETE', 'SKIP'):
                    continue

                if age_days < UNUSED_ACCESS_KEY_DAYS:
                    continue
                deactivation_grace_days = min(age_days - UNUSED_ACCESS_KEY_DAYS, days_to_take_action)
                cleanup_result = self.verify_and_delete_resource(
                    resource_id=user_name,
                    tags=tags,
                    clean_up_days=deactivation_grace_days,
                    access_key_label=access_key_label,
                )
                resource_data = self._get_es_schema(
                    resource_id=user_name,
                    user=self.get_tag_name_from_tags(tags=tags, tag_name='User'),
                    skip_policy=self.get_skip_policy_value(tags=tags),
                    cleanup_days=deactivation_grace_days,
                    dry_run=self._dry_run,
                    name=user_name,
                    region=region,
                    cleanup_result=str(cleanup_result),
                    resource_action=self.RESOURCE_ACTION,
                    cloud_name=self._cloud_name,
                    resource_type='UnusedAccessKey',
                    resource_state='Active',
                    age_days=age_days,
                    last_activity_days=last_activity_days,
                    unit_price=0,
                )
                unused_access_keys.append(resource_data)

        return unused_access_keys



from datetime import datetime
from time import strftime

import json
from pkg_resources import resource_filename

from cloud_governance.common.clouds.aws.utils.common_methods import get_boto3_client
from cloud_governance.common.logger.init_logger import logger
# Search product filter
from cloud_governance.main.environment_variables import environment_variables

FLT = '[{{"Field": "tenancy", "Value": "shared", "Type": "TERM_MATCH"}},' \
      '{{"Field": "operatingSystem", "Value": "{o}", "Type": "TERM_MATCH"}},' \
      '{{"Field": "preInstalledSw", "Value": "NA", "Type": "TERM_MATCH"}},' \
      '{{"Field": "instanceType", "Value": "{t}", "Type": "TERM_MATCH"}},' \
      '{{"Field": "location", "Value": "{r}", "Type": "TERM_MATCH"}},' \
      '{{"Field": "capacitystatus", "Value": "Used", "Type": "TERM_MATCH"}}]'

FLT_VOLUME = '[{{"Field": "volumeType", "Value": "{v}", "Type": "TERM_MATCH"}},' \
             '{{"Field": "location", "Value": "{r}", "Type": "TERM_MATCH"}} ]'


class AWSPrice:
    """
    This class return aws resource price
    """

    def __init__(self, region_name: str = ''):
        # Use AWS Pricing API at US-East-1
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__client = get_boto3_client('pricing', region_name='us-east-1')
        self.region = region_name if region_name else self.__environment_variables_dict.get('AWS_DEFAULT_REGION', 'us-east-1')

    # Get current AWS price for an on-demand instance
    def get_price(self, **kwargs):
        """
        This method return price per region
        @return:
        """
        region = kwargs.get('region')
        if kwargs.get('aws_resource') == 'ebs':
            volume_type = kwargs.get('volume_type')
            f = FLT_VOLUME.format(r=region, v=volume_type)
        else:
            instance = kwargs.get('instance')
            os = kwargs.get('os')
            f = FLT.format(r=region, t=instance, o=os)
        try:
            data = self.__client.get_products(ServiceCode='AmazonEC2', Filters=json.loads(f))
            od = json.loads(data['PriceList'][0])['terms']['OnDemand']
            id1 = list(od)[0]
            id2 = list(od[id1]['priceDimensions'])[0]
            return od[id1]['priceDimensions'][id2]['pricePerUnit']['USD']
        except Exception as err:
            return 0

    # Translate region code to region name
    def get_region_name(self, region_code):
        """
        This method return region name
        @param region_code:
        @return:
        """
        default_region = 'us-east-1'
        endpoint_file = resource_filename('botocore', 'data/endpoints.json')
        try:
            with open(endpoint_file, 'r') as f:
                data = json.load(f)
            return data['partitions'][0]['regions'][region_code]['description']
        except IOError:
            return default_region

    def get_ebs_cost(self, volume_type: str, region: str):
        """
        This method return the cost of ebs storage
        @return:
        """
        volume_types = {
            'gp': 'General Purpose',
            'io': 'Provisioned IOPS',
            'st1': 'Throughput Optimized HDD',
            'standard': 'Magnetic',
            'sc1': 'Cold HDD'
        }
        key = ''
        for volume_key, value in volume_types.items():
            if volume_key in volume_type:
                key = value
        return float(self.get_price(region=self.get_region_name(region), volume_type=key, aws_resource='ebs'))

    def get_ec2_price(self, resource: str, item_data: dict):
        """
        This method calculate ec2 cost from launch time or ebs per month in $
        @return:cluster_cost_results
        """
        if resource == 'ec2' and item_data['State']['Name'] == 'running':
            # Get current price for a given 'running' instance, region and os
            ec2_type_cost = '0'
            try:
                ec2_type_cost = self.get_price(region=self.get_region_name(self.region), instance=item_data['InstanceType'], os='Linux')
            except Exception as err:
                return 'NA'
            ec2_lanuch_time = item_data['LaunchTime']
            d1 = datetime.strptime(ec2_lanuch_time, "%Y-%m-%dT%H:%M:%S+00:00")
            d2 = datetime.strptime(strftime("%Y-%m-%dT%H:%M:%S+00:00"), "%Y-%m-%dT%H:%M:%S+00:00")
            diff = d2 - d1
            diff_in_hours = diff.total_seconds() / 3600
            ec2_cost = round(float(ec2_type_cost) * diff_in_hours, 3)
            return round(ec2_cost, 3)
        elif resource == 'ec2' and item_data['State']['Name'] != 'running':
            return '0'
        elif resource == 'ebs':
            ebs_monthly_cost = '0'
            if item_data['VolumeType'] in ('gp2', 'gp3'):
                ebs_monthly_cost = self.get_ebs_cost(volume_type='gp', region=self.region) * item_data['Size']
            elif item_data['VolumeType'] in ('io1', 'io2'):
                ebs_monthly_cost = self.get_ebs_cost(volume_type='io', region=self.region) * item_data['Size']
            elif item_data['VolumeType'] == 'standard':
                ebs_monthly_cost = self.get_ebs_cost(volume_type='standard', region=self.region) * item_data['Size']
            elif item_data['VolumeType'] == 'st1':
                ebs_monthly_cost = self.get_ebs_cost(volume_type='st1', region=self.region) * item_data['Size']
            else:
                if item_data['VolumeType'] == 'sc1':
                    ebs_monthly_cost = self.get_ebs_cost(volume_type='sc1', region=self.region) * item_data['Size']
            return round(ebs_monthly_cost, 3)

    def get_service_pricing(self, service_code: str, filter_list: list) -> float:
        """
        This method returns the price of the product by service
        :param service_code:
        :type service_code:
        :param filter_list:
        :type filter_list:
        :return:
        :rtype:
        """
        try:
            data = self.__client.get_products(ServiceCode=service_code, Filters=filter_list)
            od = json.loads(data['PriceList'][0])['terms']['OnDemand']
            id1 = list(od)[0]
            id2 = list(od[id1]['priceDimensions'])[0]
            return float(od[id1]['priceDimensions'][id2]['pricePerUnit']['USD'])
        except Exception as err:
            print(err)
            logger.error(err)
            return 0



import csv
import os.path

import google.auth
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError

from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.main.environment_variables import environment_variables


class GoogleDriveOperations:
    """
    This class perform the Google Drive Operations
    methods
    1. download_spreadsheet
    """

    RETRIES = 3

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__service = None
        if self.__environment_variables_dict.get('GOOGLE_APPLICATION_CREDENTIALS'):
            self.__creds, _ = google.auth.default()
            self.__service = build('sheets', 'v4', credentials=self.__creds, num_retries=self.RETRIES)

    @logger_time_stamp
    def create_work_sheet(self, gsheet_id: str, sheet_name: str):
        """
        This method checks for worksheet and create the worksheet
        @return:
        """
        try:
            sheet_id = self.find_sheet_id_by_name(sheet_name=sheet_name, spreadsheet_id=gsheet_id)
            if sheet_id == '':
                create_worksheet_meta_data = {
                    'requests': [{
                        'addSheet': {
                            'properties': {
                                'title': sheet_name
                            }
                        }
                    }]}
                if self.__service:
                    self.__service.spreadsheets().batchUpdate(spreadsheetId=gsheet_id,
                                                              body=create_worksheet_meta_data).execute()
                    logger.info(f'{sheet_name} worksheet created')
            else:
                logger.info(f'{sheet_name} Worksheet Already present')
        except Exception as err:
            raise err

    @logger_time_stamp
    def download_spreadsheet(self, spreadsheet_id: str, sheet_name: str, file_path: str):
        """
        This method download spreadsheet from the Google Drive
        Used the Google Drive API
        Create GCP Project, enable Google Drive API, enable Google Spreadsheet API
        Create Service Account, and generate keys in json format
        export GOOGLE_APPLICATION_CREDENTIALS=file_location
        @return:
        """
        if self.__service:
            try:
                result = self.__service.spreadsheets().values().get(spreadsheetId=spreadsheet_id,
                                                                    range=sheet_name).execute()
                file_name = f'{sheet_name}.csv'
                output_file = os.path.join(file_path, file_name)
                if result.get('values'):
                    with open(output_file, 'w') as f:
                        writer = csv.writer(f)
                        writer.writerows(result.get('values'))
                    logger.info(f'Successfully downloaded {sheet_name}.csv')
            except HttpError as error:
                logger.info(f'An error occurred: {error}')

    @logger_time_stamp
    def append_values(self, spreadsheet_id, sheet_name: str, values: list, value_input_option: str = 'USER_ENTERED'):
        """
        This method append the values in the spreadsheet
        @param spreadsheet_id:
        @param sheet_name:
        @param values:
        @param value_input_option:
        @return:
        """
        if self.__service:
            try:
                body = {'values': values}
                result = self.__service.spreadsheets().values().append(
                    spreadsheetId=spreadsheet_id, range=sheet_name,
                    valueInputOption=value_input_option, body=body).execute()
                logger.info(f'Data is append to the end of the worksheet {sheet_name}')
                return result
            except HttpError as error:
                logger.info(f'An error occurred: {error}')

    @logger_time_stamp
    def find_sheet_id_by_name(self, sheet_name: str, spreadsheet_id: str):
        """
        This method find the sheet id in the spreadsheet
        @param sheet_name:
        @param spreadsheet_id:
        @return:
        """
        if self.__service:
            sheets_with_properties = self.__service.spreadsheets().get(spreadsheetId=spreadsheet_id,
                                                                       fields='sheets.properties').execute().get(
                'sheets')
            for sheet in sheets_with_properties:
                if 'title' in sheet['properties'].keys():
                    if sheet['properties']['title'] == sheet_name:
                        return sheet['properties']['sheetId']
        return ''

    @logger_time_stamp
    def delete_rows(self, spreadsheet_id: str, sheet_name: str, row_number: int):
        """
        This method delete row from the spreadsheet bases on row number
        @param spreadsheet_id:
        @param sheet_name:
        @param row_number:
        @return:
        """
        if self.__service:
            try:
                spreadsheet_data = [
                    {
                        "deleteDimension": {
                            "range": {
                                "sheetId": self.find_sheet_id_by_name(sheet_name=sheet_name,
                                                                      spreadsheet_id=spreadsheet_id),
                                "dimension": "ROWS",
                                "startIndex": row_number,
                                "endIndex": row_number + 1
                            }
                        }
                    }
                ]
                update_data = {"requests": spreadsheet_data}
                updating = self.__service.spreadsheets().batchUpdate(spreadsheetId=spreadsheet_id, body=update_data)
                updating.execute()
            except HttpError as error:
                logger.into(f'An error occurred: {error}')

    @logger_time_stamp
    def paste_csv_to_gsheet(self, csv_path, spreadsheet_id: str, sheet_name: str):
        """
        This method paste the csv data into the specific sheet
        Note: replaces the content in the rowIndex, columnIndex you specified
        @param csv_path:
        @param spreadsheet_id:
        @param sheet_name:
        @return:
        """
        if self.__service:
            csv_contents = None
            with open(csv_path, 'r') as csv_file:
                csv_contents = csv_file.read()
            if csv_contents:
                sheet_id = self.find_sheet_id_by_name(sheet_name=sheet_name, spreadsheet_id=spreadsheet_id)
                body = {
                    'requests': [{
                        'pasteData': {
                            "coordinate": {
                                "sheetId": sheet_id,
                                "rowIndex": "0",
                                "columnIndex": "0",
                            },
                            "data": csv_contents,
                            "type": 'PASTE_NORMAL',
                            "delimiter": ',',
                        }
                    }]
                }
                request = self.__service.spreadsheets().batchUpdate(spreadsheetId=spreadsheet_id, body=body)
                response = request.execute()
                logger.info(f'Pasted data into the {sheet_name}')
                return response

    @logger_time_stamp
    def update_row_in_gsheet(self, data: list, gsheet_id: str, row: int, sheet_name: str):
        """
        This method update the values in a row
        @return:
        """
        if self.__service:
            sheet_id = self.find_sheet_id_by_name(sheet_name=sheet_name, spreadsheet_id=gsheet_id)
            requests_body = {'requests': [{
                'updateCells': {
                    'rows': [{"values": data}],
                    'fields': '*',
                    'start': {
                        "sheetId": sheet_id,
                        "rowIndex": row,
                        "columnIndex": '0'
                    }
                }
            }]}
            request = self.__service.spreadsheets().batchUpdate(spreadsheetId=gsheet_id, body=requests_body)
            response = request.execute()
            logger.info(f'Updated the row in the worksheet {sheet_name}')
            return response



import boto3

from cloud_governance.common.logger.init_logger import logger


class SupportOperations:
    """
    This class performs the support operations
    """

    def __init__(self):
        self.__client = boto3.client('support', region_name='us-east-1')

    def get_describe_trusted_advisor_checks(self):
        """
        This method returns the trusted advisor check results
        :return:
        :rtype:
        """
        try:
            response = self.__client.describe_trusted_advisor_checks(language='en')
            return response.get('checks', [])
        except Exception as err:
            logger.error(err)
        return []

    def get_trusted_advisor_reports(self):
        """
        This method returns the reports of the checks
        :return:
        :rtype:
        """
        result = {}
        try:
            advisor_checks_list = self.get_describe_trusted_advisor_checks()
            for check in advisor_checks_list:
                response = self.__client.describe_trusted_advisor_check_result(checkId=check.get('id'))
                result.setdefault(check.get('category'), {}).setdefault(check.get('id'), {
                    'metadata': check,
                    'reports': response.get('result', [])
                })
        except Exception as err:
            logger.err(err)
        return result



import time
from functools import wraps
from urllib.parse import urlparse, parse_qs

import ibm_vpc
from typing import Callable

from cloud_governance.common.clouds.ibm.account.ibm_authenticator import IBMAuthenticator


def region_wrapper(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        vpc_obj = VpcInfraOperations()
        regions = vpc_obj.get_regions()
        resources_list = {}
        exec_func = getattr(vpc_obj, func(*args, **kwargs), None)
        for region in regions:
            region_name = region.get('name')
            if region['status'] == 'available':
                vpc_obj.set_service_url(region_name)
                result = exec_func() if exec_func else []
                if result:
                    owned_resources = []
                    for resource in result:
                        if 'crn' in resource:
                            if vpc_obj.account_id in resource['crn']:
                                owned_resources.append(resource)
                        else:
                            owned_resources.append(resource)
                    if owned_resources:
                        resources_list[region_name] = owned_resources
        return resources_list

    return wrapper


class VpcInfraOperations(IBMAuthenticator):
    """
    This class contains methods to perform operations on VPC Infra Operations.
    """

    REGION_SERVICE_URL = "https://%s.iaas.cloud.ibm.com/v1"

    def __init__(self):
        super().__init__()
        self.__client = ibm_vpc.vpc_v1.VpcV1(authenticator=self.iam_authenticator)

    def get_regions(self):
        """
        This method lists all available regions.
        :return:
        """
        regions = self.__client.list_regions().get_result()['regions']
        return regions

    def set_service_url(self, region_name: str):
        """
        This method sets the service URL.
        :param region_name:
        :return:
        """
        service_url = self.REGION_SERVICE_URL % region_name
        self.__client.set_service_url(service_url)

    def iter_next_resources(self, exec_func: Callable, resource_name: str, region_name: str = None, **kwargs):
        """
        This method .
        :param region_name:
        :param exec_func:
        :param resource_name:
        :return:
        """
        if region_name:
            self.set_service_url(region_name)
        response = exec_func(**kwargs).get_result()
        resources = response[resource_name]
        count = 1
        while response.get('next'):
            parsed_url = urlparse(response['next']['href'])
            params = parse_qs(parsed_url.query)
            if params and 'start' in params:
                start = params['start'][0]
                response = exec_func(start=start, **kwargs).get_result()
                resources.extend(response[resource_name])
            count += 1
            if count == 5:
                time.sleep(30)
                count = 0
        return resources

    def get_instances(self, region_name: str = None):
        """
        This method lists available instances in one region, default 'us-south'
        :param region_name:
        :return:
        """
        return self.iter_next_resources(exec_func=self.__client.list_instances,
                                        resource_name='instances', region_name=region_name)

    def get_images(self, region_name: str = None):
        """
        This method lists available images in one region, default 'us-south'
        :param region_name:
        :return:
        """
        return self.iter_next_resources(exec_func=self.__client.list_images, resource_name='images',
                                        region_name=region_name, status='available')

    def get_placement_groups(self, region_name: str = None):
        """
        This method returns available placement groups in one region, default 'us-south'
        :param region_name:
        :return:
        """
        return self.iter_next_resources(exec_func=self.__client.list_placement_groups,
                                        resource_name='placement_groups', region_name=region_name)

    def get_volumes(self, region_name: str = None):
        """
        This method lists available volumes.
        :return:
        """
        return self.iter_next_resources(exec_func=self.__client.list_volumes,
                                        resource_name='volumes', region_name=region_name)

    def get_floating_ips(self, region_name: str = None):
        """
        This method lists available floating ips.
        :param region_name:
        :return:
        """
        return self.iter_next_resources(exec_func=self.__client.list_floating_ips,
                                        resource_name='floating_ips', region_name=region_name)

    def get_vpcs(self, region_name: str = None):
        """
        This method lists available vpcs.
        :param region_name:
        :return:
        """
        return self.iter_next_resources(exec_func=self.__client.list_vpcs,
                                        resource_name='vpcs', region_name=region_name)

    def get_virtual_network_interfaces(self, region_name: str = None):
        """
        This method lists available virtual network interfaces.
        :param region_name:
        :return:
        """
        return self.iter_next_resources(exec_func=self.__client.list_virtual_network_interfaces,
                                        resource_name='virtual_network_interfaces', region_name=region_name)

    def get_security_groups(self, region_name: str = None):
        """
        This method lists available security_groups
        :param region_name:
        :return:
        """
        return self.iter_next_resources(exec_func=self.__client.list_security_groups,
                                        resource_name='security_groups', region_name=region_name)

    def get_public_gateways(self, region_name: str = None):
        """
        This method lists available public_gateways
        :param region_name:
        :return:
        """
        return self.iter_next_resources(exec_func=self.__client.list_public_gateways,
                                        resource_name='public_gateways', region_name=region_name)

    def get_vpc_endpoint_gateways(self, region_name: str = None):
        """
        This method lists available vpc endpoint gateways
        :param region_name:
        :return:
        """
        return self.iter_next_resources(exec_func=self.__client.list_endpoint_gateways,
                                        resource_name='endpoint_gateways', region_name=region_name)

    def get_load_balancers(self, region_name: str = None):
        """
        This method lists available load balancers
        :param region_name:
        :return:
        """
        return self.iter_next_resources(exec_func=self.__client.list_load_balancers,
                                        resource_name='load_balancers', region_name=region_name)

    def get_baremetal_servers(self, region_name: str = None):
        """
        This method lists available baremetals
        :param region_name:
        :return:
        """
        return self.iter_next_resources(exec_func=self.__client.list_bare_metal_servers,
                                        resource_name='bare_metal_servers', region_name=region_name)

    @region_wrapper
    def get_all_instances(self):
        """
        This method lists all available instances.
        :return:
        """
        return "get_instances"

    @region_wrapper
    def get_all_volumes(self):
        """
        This method lists all available volumes.
        :return:
        """
        return "get_volumes"

    @region_wrapper
    def get_all_vpcs(self):
        """
        This method lists all available vpc's.
        :return:
        """
        return "get_vpcs"

    @region_wrapper
    def get_all_floating_ips(self):
        """
        This method lists all floating ips.
        :return:
        """
        return "get_floating_ips"

    @region_wrapper
    def get_all_virtual_network_interfaces(self):
        """
        This method lists all available virtual network interfaces.
        :return:
        """
        return "get_virtual_network_interfaces"

    @region_wrapper
    def get_all_security_groups(self):
        """
        This method lists all available security_groups
        :return:
        """
        return "get_security_groups"

    @region_wrapper
    def get_all_public_gateways(self):
        """
        This method lists all available public_gateways
        :return:
        """
        return "get_public_gateways"

    @region_wrapper
    def get_all_vpc_endpoint_gateways(self):
        """
        This method lists all available vpc endpoint gateways
        :return:
        """
        return "get_vpc_endpoint_gateways"

    @region_wrapper
    def get_all_load_balancers(self):
        """
        This method lists all available load balancers.
        :return:
        """
        return "get_load_balancers"

    @region_wrapper
    def get_all_baremetal_servers(self):
        """
        This method lists all available baremetals.
        :return:
        """
        return "get_baremetal_servers"

    @region_wrapper
    def get_all_placement_groups(self):
        """
        This method lists all available placement groups.
        :return:
        """
        return "get_placement_groups"

    @region_wrapper
    def get_all_images(self):
        """
        This method lists all available images.
        :return:
        """
        return "get_images"



import json

from cloud_governance.common.utils.api_requests import APIRequests
from cloud_governance.common.utils.configs import LOOK_BACK_DAYS
from cloud_governance.main.environment_variables import environment_variables
from datetime import datetime, timedelta, timezone


class CloudabilityOperations:
    APPITO_LOGIN_API = "https://frontdoor.apptio.com/service/apikeylogin"

    def __init__(self):
        self.__api_requests = APIRequests()
        self.environment_variables_dict = environment_variables.environment_variables_dict
        self.__view_id = self.environment_variables_dict.get('CLOUDABILITY_VIEW_ID')
        self.__cloudability_api = self.environment_variables_dict.get('CLOUDABILITY_API')
        self.__reports_path = self.environment_variables_dict.get('CLOUDABILITY_API_REPORTS_PATH')
        self.__appito_envid = self.environment_variables_dict.get('APPITO_ENVID')
        self.__key_secret = self.environment_variables_dict.get('APPITO_KEY_SECRET')
        self.__key_access = self.environment_variables_dict.get('APPITO_KEY_ACCESS')
        self.dimensions = self.environment_variables_dict.get('CLOUDABILITY_DIMENSIONS')
        self.metrics = self.environment_variables_dict.get('CLOUDABILITY_METRICS')

    def __get_appito_token(self):
        """
        This method returns the appito token
        :return:
        :rtype:
        """
        data = {
            "keyAccess": self.__key_access,
            "keySecret": self.__key_secret
        }
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        response = self.__api_requests.post(url=self.APPITO_LOGIN_API, data=json.dumps(data), headers=headers)
        if response.ok:
            return response.headers['apptio-opentoken']
        return None

    def __get_start_date(self, look_back_days: int = LOOK_BACK_DAYS):
        return (self.__get_end_date() - timedelta(days=look_back_days)).replace(day=1)

    def __get_end_date(self):
        return datetime.now(timezone.utc).date()

    def get_cost_reports(self,
                         dimensions: str = None,
                         metrics: str = None,
                         start_date: str = None,
                         end_date: str = None,
                         custom_filter: str = '',
                         look_back_days: int = LOOK_BACK_DAYS,
                         page_token: str = None):
        """
        This method returns the cost reports from the cloudability
        :return:
        :rtype:
        """
        cost_usage_result = []
        appito_token = self.__get_appito_token()
        if not appito_token:
            raise Exception("Appito Token missing error")

        dimensions = dimensions or self.dimensions
        metrics = metrics or self.metrics
        start_date = start_date or self.__get_start_date(look_back_days)
        end_date = end_date or self.__get_end_date()
        api = f'{self.__cloudability_api}/{self.__reports_path}?' \
              f'dimensions={dimensions}&metrics={metrics}' \
              f'&start_date={start_date}&end_date={end_date}' \
              f'&id={self.__view_id}&{custom_filter}'
        if page_token:
            api += f'&token={page_token}'
        headers = {
            "apptio-environmentid": self.__appito_envid,
            "apptio-opentoken": appito_token
        }
        response = self.__api_requests.get(url=api, headers=headers)
        if isinstance(response, dict):
            results = response.get('results') if response.get('results') else response
            cost_usage_result.extend(results)
            page_token = response.get('pagination', {}).get('next')
            if page_token:
                cost_usage_result.extend(self.get_cost_reports(page_token=page_token,
                                                               dimensions=dimensions,
                                                               metrics=metrics,
                                                               start_date=start_date,
                                                               end_date=end_date,
                                                               custom_filter=custom_filter
                                                               ))

        return cost_usage_result



from abc import ABC, abstractmethod
from typing import Union

from cloud_governance.main.environment_variables import environment_variables


class AbstractUpload(ABC):

    def __init__(self):
        self._environment_variables_dict = environment_variables.environment_variables_dict
        self._account = self._environment_variables_dict.get('account', '')
        self._region = self._environment_variables_dict.get('AWS_DEFAULT_REGION', 'us-east-2')
        self._es_index = self._environment_variables_dict.get('es_index')
        self._policy_output = self._environment_variables_dict.get('policy_output', '')
        self._policy = self._environment_variables_dict.get('policy', '')

    @abstractmethod
    def upload(self, data: Union[list, dict]):
        """
        This method upload data
        :return:
        :rtype:
        """
        raise NotImplemented("This is not yet implemented")



from datetime import datetime, timedelta

from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.utils.configs import INSTANCE_START_PREFIX
from cloud_governance.policy.policy_operations.aws.tag_non_cluster.non_cluster_operations import NonClusterOperations


class TagNonClusterResources(NonClusterOperations):
    """
    This class update tags of Non-Cluster Resources
    """

    SHORT_RESOURCE_ID = 5
    SHORT_RESOURCE_NAME = 3
    SHOT_SNAPSHOT_ID = 4

    def __init__(self, region: str = 'us-east-2', dry_run: str = 'yes', input_tags: dict = ''):
        super().__init__(region=region, dry_run=dry_run, input_tags=input_tags)

    def __check_name_in_tags(self, tags: list):
        """
        This method check name in tags
        @param tags:
        @return:
        """
        if tags:
            for tag in tags:
                if tag.get('Key') == 'Name':
                    return True
        return False

    def __get_instance_tags(self, launch_time: datetime, instance_id: str, tags: list):
        """
        This method returns the tags to update  the instance tags
        @param launch_time:
        @param instance_id:
        @param tags:
        @return:
        """
        username = self.get_username(start_time=launch_time, resource_id=instance_id,
                                     resource_type='AWS::EC2::Instance', tags=tags)
        search_tags = []
        user_tags = []
        if not username:
            search_tags.extend(self._fill_na_tags())
        else:
            search_tags.append(self._build_tag(key='Email', value=f'{username}@redhat.com'))
            user_tags = self.iam_client.get_user_tags(username=username)
            if not user_tags:
                search_tags.extend(self._fill_na_tags(user=username))
        if not self.__check_name_in_tags(tags):
            tag_name = f'{username}-{instance_id[-self.SHORT_RESOURCE_ID:]}' if username else f'{instance_id[0:1]}-{self.region}-{instance_id[-self.SHORT_RESOURCE_ID:]}'
            search_tags.append(self._build_tag(key='cg-Name', value=tag_name))
        search_tags.append(self._build_tag(key='LaunchTime', value=launch_time))
        search_tags.extend(self._append_input_tags())
        search_tags.extend(user_tags)
        add_tags = self._get_tags_of_resources(tags=tags, search_tags=search_tags)
        return add_tags

    def non_cluster_update_ec2(self, instances_list: list = None):
        """
        This method tagged the ec2 instances without having tags
        @return:
        """
        if not instances_list:
            instances_list = self._get_resource_data(resource_method=self._get_instances_data)
        instances_ids = []
        for instance in instances_list:
            for item in instance:
                instance_id = item.get('InstanceId')
                launch_time = item.get('LaunchTime')
                tags = item.get('Tags')
                if not self.validate_existing_tag(tags=tags):
                    add_tags = self.__get_instance_tags(launch_time=launch_time, instance_id=instance_id, tags=tags)
                    if add_tags:
                        if self.dry_run == 'no':
                            try:
                                self.utils.tag_aws_resources(client_method=self.ec2_client.create_tags,
                                                             resource_ids=[instance_id], tags=add_tags)
                                logger.info(
                                    f'Added tags to instance: {instance_id} total: {len(add_tags)} tags: {add_tags}')
                            except Exception as err:
                                logger.info(err)
                        instances_ids.append(instance_id)
        logger.info(f'non_cluster_ec2 count: {len(sorted(instances_ids))} {sorted(instances_ids)}')
        return sorted(instances_ids)

    def update_volumes(self, volumes_data: list = None):
        """
        This method updates the tags of non-cluster volumes
        @param volumes_data:
        @return:
        """
        if not volumes_data:
            volumes_data = self._get_resource_data(resource_method=self.ec2_operations.get_volumes)
        volume_ids = []
        for volume in volumes_data:
            volume_id = volume.get('VolumeId')
            tags = volume.get('Tags')
            if not self.validate_existing_tag(tags=tags):
                username = self.get_username(start_time=volume.get('CreateTime'), resource_id=volume_id,
                                             resource_type='AWS::EC2::Volume', tags=tags)
                search_tags = []
                if not username:
                    get_tags, username = self._get_tags_fom_attachments(attachments=volume.get('Attachments'))
                    search_tags.extend(get_tags)
                else:
                    search_tags.extend(self._append_input_tags())
                if username:
                    user_tags = self.iam_client.get_user_tags(username=username)
                    if not user_tags:
                        search_tags.extend(self._fill_na_tags(user=username))
                    else:
                        search_tags.extend(user_tags)
                        search_tags.append({'Key': 'Email', 'Value': f'{username}@redhat.com'})
                    search_tags.append(self._build_tag(key='LaunchTime', value=volume.get('CreateTime')))
                else:
                    search_tags.extend(self._fill_na_tags())
                    search_tags.extend(self._append_input_tags())
                    search_tags.append(self._build_tag(key='LaunchTime', value=volume.get('CreateTime')))
                if not self.__check_name_in_tags(volume.get('Tags')):
                    tag_name = f'{username}-{volume_id[-self.SHORT_RESOURCE_ID:]}' if username else f'{volume_id[:self.SHORT_RESOURCE_NAME]}-{self.region}-{volume_id[-self.SHORT_RESOURCE_ID:]}'
                    search_tags.append({'Key': 'cg-Name', 'Value': tag_name})
                volume_tags = self._get_tags_of_resources(tags=volume.get('Tags'), search_tags=search_tags)
                if volume_tags:
                    if self.dry_run == 'no':
                        try:
                            self.utils.tag_aws_resources(client_method=self.ec2_client.create_tags,
                                                         resource_ids=[volume_id], tags=volume_tags)
                            logger.info(
                                f'added tags to volume_id: {volume_id} total: {len(volume_tags)}  tags: {volume_tags}')
                        except Exception as err:
                            logger.info(err)
                    volume_ids.append(volume_id)
        logger.info(f'non_cluster_volumes count: {len(sorted(volume_ids))} {sorted(volume_ids)}')
        return sorted(volume_ids)

    def update_snapshots(self, snapshots: list = None):
        """
        This method updates the tags of  non-cluster snapshots
        @param snapshots:
        @return:
        """
        if not snapshots:
            snapshots = self._get_resource_data(resource_method=self.ec2_operations.get_snapshots)
        snapshot_ids = []
        for snapshot in snapshots:
            snapshot_id = snapshot.get('SnapshotId')
            tags = snapshot.get('Tags')
            if not self.validate_existing_tag(tags=tags):
                username = self.get_username(start_time=snapshot.get('StartTime'), resource_id=snapshot_id,
                                             resource_type='AWS::EC2::Snapshot', tags=tags)
                if 'vm_import_image' in username:
                    start_time = snapshot.get('StartTime') + timedelta(seconds=5)
                    end_time = start_time + timedelta(minutes=30)
                    assume_username = self.get_username(start_time=start_time, resource_id=snapshot_id,
                                                        resource_type='AWS::EC2::Snapshot', tags=tags,
                                                        end_time=end_time)
                    if assume_username:
                        username = assume_username
                search_tags = []
                if not username:
                    if snapshot.get('Description') and 'Created' in snapshot.get('Description'):
                        image_tags, username = self._get_tags_from_snapshot_description_images(
                            description=snapshot.get('Description'))
                        if not username:
                            if INSTANCE_START_PREFIX in snapshot.get('Description'):
                                try:
                                    instance_id = snapshot.get('Description').split(" ")[2].split("(")[1][:-1]
                                    instances = self._get_instances_data(instance_id)
                                    if instances:
                                        for item in instances:
                                            if item.get('InstanceId') == instance_id:
                                                item_tags, username = self._get_tags_from_instance_item(
                                                    instance_item=item)
                                except Exception as err:
                                    logger.info(err)
                else:
                    search_tags.extend(self._append_input_tags())
                if username:
                    user_tags = self.iam_client.get_user_tags(username=username)
                    search_tags.append({'Key': 'Email', 'Value': f'{username}@redhat.com'})
                    if not user_tags:
                        search_tags.extend(self._fill_na_tags(user=username))
                    else:
                        search_tags.extend(user_tags)
                else:
                    search_tags.extend(self._fill_na_tags())
                    search_tags.extend(self._append_input_tags())
                if not self.__check_name_in_tags(snapshot.get('Tags')):
                    tag_name = f'{username}-{snapshot_id[-self.SHORT_RESOURCE_ID:]}' if username else f'{snapshot_id[:self.SHOT_SNAPSHOT_ID]}-{self.region}-{snapshot_id[-self.SHORT_RESOURCE_ID:]}'
                    search_tags.append({'Key': 'cg-Name', 'Value': tag_name})
                search_tags.append(self._build_tag(key='LaunchTime', value=snapshot.get('StartTime')))
                snapshot_tags = self._get_tags_of_resources(tags=snapshot.get('Tags'), search_tags=search_tags)
                if snapshot_tags:
                    if self.dry_run == 'no':
                        try:
                            self.utils.tag_aws_resources(client_method=self.ec2_client.create_tags,
                                                         resource_ids=[snapshot_id], tags=snapshot_tags)
                            logger.info(
                                f'added tags to snapshots: {snapshot_id} total: {len(snapshot_tags)} tags: {snapshot_tags}')
                        except Exception as err:
                            logger.info(err)
                    snapshot_ids.append(snapshot_id)
        logger.info(f'non_cluster_snapshot count: {len(sorted(snapshot_ids))} {sorted(snapshot_ids)}')
        return sorted(snapshot_ids)

    def update_ami(self, images: list = None):
        """
        This method update the tags of non-cluster Amazon Machine Images
        @param images:
        @return:
        """
        if not images:
            images = self.ec2_operations.get_images()
            _, images = self.ec2_operations.scan_cluster_non_cluster_resources(images)
        image_ids = []
        for image in images:
            image_id = image.get('ImageId')
            tags = image.get('Tags')
            image_name = image.get('Name')
            start_time = datetime.fromisoformat(image.get('CreationDate')[:-1] + '+00:00')
            if not self.validate_existing_tag(tags=tags):
                username = self.get_username(start_time=start_time, resource_id=image_id, resource_type='AWS::EC2::Ami',
                                             tags=tags, resource_name=image_name)
                search_tags = []
                search_tags.extend(self._append_input_tags())
                if username:
                    user_tags = self.iam_client.get_user_tags(username=username)
                    search_tags.append({'Key': 'Email', 'Value': f'{username}@redhat.com'})
                    if not user_tags:
                        search_tags.extend(self._fill_na_tags(user=username))
                    else:
                        search_tags.extend(user_tags)
                else:
                    search_tags.extend(self._fill_na_tags())
                if not self.__check_name_in_tags(image.get('Tags')):
                    tag_name = f'{username}-{image_id[-self.SHORT_RESOURCE_ID:]}' if username else f'{image_id[:self.SHORT_RESOURCE_NAME]}-{self.region}-{image_id[-self.SHORT_RESOURCE_ID:]}'
                    search_tags.append({'Key': 'cg-Name', 'Value': tag_name})
                search_tags.append(self._build_tag(key='LaunchTime', value=start_time))
                image_tags = self._get_tags_of_resources(tags=image.get('Tags'), search_tags=search_tags)
                if image_tags:
                    if self.dry_run == 'no':
                        try:
                            self.utils.tag_aws_resources(client_method=self.ec2_client.create_tags,
                                                         resource_ids=[image_id], tags=image_tags)
                            logger.info(f'added tags to image: {image_id} total: {len(image_tags)} tags: {image_tags}')
                        except Exception as err:
                            logger.info(err)
                    image_ids.append(image_id)
        logger.info(f'non_cluster_amis count: {len(sorted(image_ids))} {sorted(image_ids)}')
        return sorted(image_ids)



from cloud_governance.common.logger.init_logger import logger
from cloud_governance.policy.helpers.aws.aws_policy_operations import AWSPolicyOperations


class EmptyRoles(AWSPolicyOperations):
    """
    This class sends an alert mail for empty role to the user after 4 days and delete after 7 days.
    """

    RESOURCE_ACTION = 'Delete'
    IAM_GLOBAL_REGION = 'us-east-1'

    def __init__(self):
        super().__init__()

    def run_policy_operations(self):
        """
        This method returns all Empty roles.
        :return:
        :rtype:
        """
        empty_roles = []
        roles = self._iam_operations.get_roles()
        for role in roles:
            role_name = role.get('RoleName')
            role_data = self._iam_operations.get_role(role_name=role_name)
            tags = role_data.get('Tags', [])
            cleanup_result = False
            cluster_tag = self._get_cluster_tag(tags=tags)
            cleanup_days = 0
            inline_policies = self._iam_operations.list_inline_role_policies(role_name=role_name)
            attached_policies = self._iam_operations.list_attached_role_policies(role_name=role_name)
            try:
                if not cluster_tag and len(inline_policies) == 0 and len(attached_policies) == 0 and \
                    self.get_skip_policy_value(tags=tags) not in ('NOTDELETE', 'SKIP'):

                    cleanup_days = self.get_clean_up_days_count(tags=tags)
                    # Step 1: Remove role from all instance profiles if attached (for empty roles)
                    instance_profiles = self._iam_operations.get_instance_profiles_for_role(role_name=role_name)
                    detachment_success = True
                    if instance_profiles:
                        logger.info(
                            f"Role '{role_name}' is empty (no policies) but attached to instance profile(s): "
                            f"{[ip['InstanceProfileName'] for ip in instance_profiles]}. Detaching now...")

                        # Call the new method to handle detachment if dry_run = 'no'
                        if self._dry_run == 'no':
                            detachment_success = self._iam_operations.remove_role_from_instance_profiles(
                                role_name=role_name)
                            logger.warning(f"LIVE RUN: Detachment attempted. Success: {detachment_success}")
                        else:
                            logger.info(f"DRY RUN: Skipping actual detachment of role '{role_name}'.")

                        if not detachment_success:
                            # Skip deletion if detachment failed
                            logger.info(
                                f"Failed to detach role '{role_name}' from all instance profiles. Skipping deletion.")
                            self.update_resource_day_count_tag(resource_id=role_name, cleanup_days=cleanup_days,
                                                               tags=tags)
                            continue

                    # Step 2: Proceed with deletion
                    cleanup_result = self.verify_and_delete_resource(resource_id=role_name, tags=tags,
                                                                     clean_up_days=cleanup_days)
                    resource_data = self._get_es_schema(resource_id=role_name,
                                                        user=self.get_tag_name_from_tags(tags=tags, tag_name='User'),
                                                        skip_policy=self.get_skip_policy_value(tags=tags),
                                                        cleanup_days=cleanup_days,
                                                        dry_run=self._dry_run,
                                                        name=role_name,
                                                        region=self.IAM_GLOBAL_REGION,
                                                        cleanup_result=str(cleanup_result),
                                                        resource_action=self.RESOURCE_ACTION,
                                                        cloud_name=self._cloud_name,
                                                        resource_type='EmptyRole',
                                                        resource_state="Empty",
                                                        unit_price=0)
                    empty_roles.append(resource_data)

                if not cleanup_result:
                    self.update_resource_day_count_tag(resource_id=role_name, cleanup_days=cleanup_days, tags=tags)

            except Exception as e:
                logger.error(f'Exception raised while processing the empty roles operation on {role_name}, {e}')
        return empty_roles



COST_USAGE_REPORT_DIMENSIONS = {
    'resource_name': 'ResourceName',
    'lease_type': 'OfferingType',
    'usage_family': 'UsageFamily',
    'instance_category': 'InstanceCategory',
    'usage_type': 'UsageType',
    'item_description': 'ItemDescription',
    'region': 'Region',
    'service_name': 'ServiceName',
    'date': 'StartDate',
    'category4': 'CostCenter',
    'vendor_account_name': 'AccountName',
    'vendor': 'PublicCloudName',
    'vendor_account_identifier': 'AccountId',
    'tag26': 'ResourcePlanId'
}

COST_USAGE_REPORT_METRICS = {
    'unblended_cost': 'Cost'
}



import boto3

from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.utils.configs import AWS_DEFAULT_GLOBAL_REGION


def get_tag_value_from_tags(tags: list, tag_name: str, cast_type: str = 'str',
                            default_value: any = '') -> any:
    """
    This method returns the tag value inputted by tag_name
    :param tags:
    :type tags:
    :param tag_name:
    :type tag_name:
    :param cast_type:
    :type cast_type:
    :param default_value:
    :type default_value:
    :return:
    :rtype:
    """
    if tags:
        tag_name = tag_name.lower().replace("_", '').replace("-", '').strip()
        for tag in tags:
            key = tag.get('Key').lower().replace("_", '').replace("-", '').strip()
            if key == tag_name.lower():
                if cast_type:
                    if cast_type == 'int':
                        return int(tag.get('Value').split()[0].strip())
                    elif cast_type == 'float':
                        return float(tag.get('Value').strip())
                    else:
                        return str(tag.get('Value').strip())
                return tag.get('Value').strip()
    return default_value


def get_boto3_client(client: str, region_name: str = AWS_DEFAULT_GLOBAL_REGION, **kwargs):
    """
    This method initializes the aws boto3 client
    :param client:
    :type client:
    :param region_name:
    :type region_name:
    :return:
    :rtype:
    """
    client_object = None
    try:
        client_object = boto3.client(client, region_name=region_name, **kwargs)
    except Exception as err:
        logger.error(f"{client} Client Initialization error: {err}")
    return client_object



from datetime import datetime

from cloud_governance.common.clouds.aws.utils.common_methods import get_tag_value_from_tags
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.main.environment_variables import environment_variables
from cloud_governance.policy.policy_operations.aws.tag_cluster.tag_cluster_operations import TagClusterOperations

from cloud_governance.policy.policy_operations.aws.tag_non_cluster.tag_non_cluster_resources import \
    TagNonClusterResources


class TagClusterResources(TagClusterOperations):
    """
    This class filter cluster resources by cluster name, and update tags when passing input_tags
    """

    SHORT_ID = 5
    NA_VALUE = 'NA'

    def __init__(self, cluster_name: str = None, cluster_prefix: list = None, input_tags: dict = None,
                 region: str = 'us-east-2', dry_run: str = 'yes', cluster_only: bool = False):
        super().__init__(cluster_name=cluster_name, cluster_prefix=cluster_prefix, input_tags=input_tags, region=region,
                         dry_run=dry_run, cluster_only=cluster_only)
        self.cluster_key = self.__init_cluster_name()
        self.non_cluster_update = TagNonClusterResources(region=region, dry_run=dry_run, input_tags=input_tags)

    def __init_cluster_name(self):
        """
        This method find the cluster full stamp key according to user cluster name, scan instance and if not found scan security group
        i.e.: user cluster name = test , cluster stamp key =  kubernetes.io/cluster/test-jlhpd
        @return:
        """
        return self.__scan_cluster_security_groups()

    def __append_input_tags(self, current_tags: list = None):
        """
        This method append the input tags to the current tags, and return the input tags
        @param current_tags:
        @return:
        """
        input_tags = self._input_tags_list_builder()
        if current_tags:
            for current_item in current_tags:
                if self.input_tags.get(current_item['Key']):
                    for input_item in input_tags:
                        if current_item['Key'] == input_item['Key']:
                            current_item['Value'] = input_item['Value']
                else:
                    input_tags.append(current_item)
        return input_tags

    def __add_cluster_id_tag(self, tags: list):
        """
        This method add cluster_id tag if not present
        :param tags:
        :type tags:
        :return:
        :rtype:
        """
        cluster_id_tag = get_tag_value_from_tags(tags=tags, tag_name=self.CLUSTER_ID_COST_ALLOCATION_TAG)
        if not cluster_id_tag:
            cluster_name = ''
            for tag in tags:
                if tag.get('Key') == 'api.openshift.com/name':
                    cluster_name = tag['Value']
                else:
                    if not cluster_name and any(prefix in tag.get('Key', '') for prefix in self.cluster_prefix):
                        cluster_name = tag['Key']
            if cluster_name:
                # cluster_name = cluster_id
                cluster_name = cluster_name.split('/')[-1]
                tags.append({'Key': self.CLUSTER_ID_COST_ALLOCATION_TAG, 'Value': cluster_name})
        return tags

    def __check_name_in_tags(self, tags: list, resource_id: str):
        """
        This method checks Name is present in the Tags or not, if not  add Name tag
        @param tags:
        @param resource_id:
        @return:
        """
        found = False
        cluster_name = self.cluster_name
        for tag in tags:
            if tag.get('Key') == 'Name':
                found = True
                break
        for tag in tags:
            if any(prefix in tag.get('Key', '') for prefix in self.cluster_prefix):
                cluster_name = tag['Key']
                break
        if not found:
            # add cg-Name if Name tag is empty in resource tags
            value = f'{cluster_name.split("/")[-1]}-{resource_id.split("-")[0]}-{resource_id[-self.SHORT_ID:]}'
            tags.append({'Key': 'cg-Name', 'Value': value})
        return tags

    def __get_cluster_tags_by_instance_cluster(self, cluster_name: str):
        """
        This method get the cluster instance tags by cluster name
        @param cluster_name:
        @return:
        """
        instances_list = self._get_instances_data()
        if instances_list:
            for instance in instances_list:
                for item in instance:
                    if item.get('Tags'):
                        for tag in item.get('Tags'):
                            if any(prefix in tag.get('Key', '') for prefix in self.cluster_prefix):
                                if cluster_name in tag.get('Key'):
                                    # Propagation-blocked prefixes derived from CLUSTER_PREFIX (e.g. kubernetes.io/, sigs.k8s.io/)
                                    cluster_prefix = environment_variables.environment_variables_dict['CLUSTER_PREFIX']
                                    no_propagate_prefixes = tuple(p.split('/', 1)[0] + '/' for p in cluster_prefix)
                                    i_tags = [instance_tag for instance_tag in item.get('Tags') if
                                              instance_tag.get('Key') != 'Name' and
                                              not any((instance_tag.get('Key') or '').startswith(prefix)
                                                      for prefix in self.cluster_prefix) and
                                              not (instance_tag.get('Key') or '').startswith(no_propagate_prefixes)]
                                    return i_tags
        return []

    def get_date_from_date(self, date_time: datetime):
        return date_time.strftime('%Y/%m/%d')

    def get_date_from_date_time(self, date_time: datetime):
        return date_time.strftime('%Y/%m/%d %H:%M:%S')

    def __remove_tags_start_with_aws(self, tags: list):
        """
        This method removes tags starting with aws
        @param tags:
        @return:
        """
        filter_tags = []
        for tag in tags:
            if not tag.get('Key').startswith('aws'):
                filter_tags.append(tag)
        return filter_tags

    def __generate_cluster_resources_list_by_tag(self, resources_list: list, input_resource_id: str,
                                                 tags: str = 'Tags'):
        """
        This method returns resource list that related to input resource id according to cluster's tag name and update the tags
        @param resources_list:
        @param input_resource_id:
        @param ids:
        @param tags:
        @return:
        """
        cluster_ids = {}
        cluster_tags = {}
        for resource in resources_list:
            resource_id = resource[input_resource_id]
            if resource.get(tags):
                # search that not exist permanent tags in the resource
                if not self.__validate_existing_tag(resource.get(tags)):
                    for tag in resource[tags]:
                        if any(prefix in tag.get('Key', '') for prefix in self.cluster_prefix):
                            if tag.get('Key') not in cluster_tags:
                                cluster_tags[tag.get('Key')] = []
                                cluster_ids[tag.get('Key')] = []
                                add_tags = self.__append_input_tags(resource.get(tags))
                                instance_tags = self.__get_cluster_tags_by_instance_cluster(cluster_name=tag.get('Key'))
                                add_tags.extend(instance_tags)
                                add_tags = self.__check_name_in_tags(tags=add_tags, resource_id=resource_id)
                                add_tags = self.__remove_tags_start_with_aws(add_tags)
                                add_tags = self.__filter_resource_tags_by_add_tags(resource.get(tags), add_tags)
                                if add_tags:
                                    cluster_tags[tag.get('Key')].extend(add_tags)
                                    cluster_ids[tag.get('Key')].append(resource_id)
                            else:
                                cluster_ids[tag.get('Key')].append(resource_id)
        result_resources_list = []
        for cluster_name, cluster_id in cluster_ids.items():
            if self.dry_run == 'no':
                try:
                    self.utils.tag_aws_resources(client_method=self.ec2_client.create_tags, resource_ids=cluster_id,
                                                 tags=cluster_tags.get(cluster_name))
                    logger.info(
                        f'{input_resource_id}:: {cluster_name}, count: {len(cluster_id)}, {cluster_id},  {cluster_tags.get(cluster_name)}')
                except Exception as err:
                    logger.info(err)
            result_resources_list.extend(cluster_id)
        return sorted(result_resources_list)

    def __generate_cluster_resources_list_by_vpc(self, resources_list: list, input_resource_id: str):
        """
        This method returns resource list that related to input resource id according to cluster's vpc id
        @param resources_list:
        @param input_resource_id:
        @return:
        """
        result_resources_list = []
        vpc_data = self.get_cluster_vpc()
        for resource in resources_list:
            resource_id = resource[input_resource_id]
            if resource.get('VpcId'):
                for vpc_id in vpc_data.keys():
                    if resource.get('VpcId') == vpc_id:
                        all_tags = []
                        all_tags.extend(vpc_data.get(vpc_id))
                        all_tags = self.__check_name_in_tags(tags=all_tags, resource_id=resource_id)
                        all_tags = self.__filter_resource_tags_by_add_tags(resource.get('Tags'), all_tags)
                        cluster_tag = [tag for tag in vpc_data.get(vpc_id, []) if any(prefix in tag.get('Key', '') for prefix in self.cluster_prefix)]
                        if all_tags:
                            if self.cluster_name:
                                if self.cluster_name in cluster_tag[0].get('Key'):
                                    if self.dry_run == 'no':
                                        self.utils.tag_aws_resources(client_method=self.ec2_client.create_tags,
                                                                     resource_ids=[resource_id], tags=all_tags)
                                        logger.info(all_tags)
                                    result_resources_list.append(resource_id)
                            else:
                                if self.dry_run == 'no':
                                    self.utils.tag_aws_resources(client_method=self.ec2_client.create_tags,
                                                                 resource_ids=[resource_id], tags=all_tags)
                                    logger.info(all_tags)
                                result_resources_list.append(resource_id)
                        break
        return sorted(result_resources_list)

    def __scan_resource_for_cluster_fullname(self, resources_list: list, tags: str = 'Tags'):
        """
        This method scan for full cluster name according in input resource by input cluster name.
        @param resources_list:
        @param tags:
        @return:
        """
        if self.cluster_name:
            for resource in resources_list:
                if resource.get(tags):
                    for tag in resource[tags]:
                        if any(tag['Key'].startswith(f"{prefix}/{self.cluster_name}") for prefix in self.cluster_prefix):
                            return tag['Key']
        return ''

    def __scan_cluster_security_groups(self):
        """
        This method scan for cluster stamp key in instances and security group, if not found return empty string
        @return:
        """
        security_groups = self.__get_security_group_data()
        # scan security group for cluster stamp key
        return self.__scan_resource_for_cluster_fullname(resources_list=security_groups)

    def remove_creation_date(self, tags: list):
        return [tag for tag in tags if tag.get('Key') != 'CreationDate']

    def __check_user_in_username_tags(self, tags: list):
        """
        This method check User tag in username tags
        @param tags:
        @return:
        """
        for tag in tags:
            if tag.get('Key') == 'User':
                return True
        return False

    def __validate_existing_tag(self, tags: list):
        """
        This method validates that permanent tag exists in tags list
        @param tags:
        @return:
        """
        check_tags = ['User', 'Project', 'Manager', 'Owner', 'Email', self.CLUSTER_ID_COST_ALLOCATION_TAG]
        tag_count = 0
        for tag in tags:
            if tag.get('Key') in check_tags:
                tag_count += 1
                if tag.get('Value') == 'NA':
                    return False
        return tag_count == len(check_tags)

    def update_cluster_tags(self, resources: list):
        """
        This method update the Cluster instance tags and returns the updated tags list ids.
        @param resources:
        @param queue:
        @return:
        """
        cluster_instances = {}
        result_instance_list = []
        cluster_tags = {}
        for instance in resources:
            for item in instance:
                instance_id = item['InstanceId']
                tags = item.get('Tags')
                if tags:
                    # search that not exist permanent tags in the resource
                    if not self.__validate_existing_tag(tags):
                        for tag in tags:
                            if any(prefix in tag.get('Key', '') for prefix in self.cluster_prefix):
                                add_tags = self.__append_input_tags()
                                add_tags.append(tag)
                                cluster_name = tag.get('Key').split('/')[-1]
                                user = self.ec2_operations.get_tag_value_from_tags(tags=tags, tag_name='User')
                                if cluster_name in cluster_instances and user and user != 'NA':
                                    add_tags = self.__filter_resource_tags_by_add_tags(tags=tags,
                                                                                       search_tags=cluster_tags[
                                                                                           cluster_name])
                                    if add_tags:
                                        cluster_instances[cluster_name].append(instance_id)
                                    break
                                else:
                                    username = self.get_username(start_time=item.get('LaunchTime'),
                                                                 resource_id=instance_id,
                                                                 resource_type='AWS::EC2::Instance', tags=tags)
                                    if username:
                                        if username == 'AutoScaling':
                                            add_tags.extend(self._fill_na_tags(user=username))
                                            logger.info(f'Autoscaling instance :: {instance_id}')
                                        else:
                                            user_tags = self.iam_operations.get_user_tags(username=username)
                                            if not self.__check_user_in_username_tags(user_tags):
                                                try:
                                                    user = self.iam_client.get_user(UserName=username)['User']
                                                    temp_username = self.cloudtrail.get_username_by_instance_id_and_time(
                                                        start_time=user.get('CreateDate'), resource_id=username,
                                                        resource_type='AWS::IAM::User')
                                                    if temp_username:
                                                        add_tags.append({'Key': 'User', 'Value': temp_username})
                                                        user_tags = self.iam_operations.get_user_tags(
                                                            username=temp_username)
                                                    else:
                                                        add_tags.append({'Key': 'User', 'Value': username})
                                                except:
                                                    add_tags.append({'Key': 'User', 'Value': username})
                                            if user_tags:
                                                add_tags.extend(user_tags)
                                                add_tags.append({'Key': 'Email', 'Value': f'{username}@redhat.com'})
                                            else:
                                                add_tags.extend(self._fill_na_tags(user=username))
                                    else:
                                        add_tags.extend(self._fill_na_tags())
                                    add_tags.append({'Key': 'LaunchTime',
                                                     'Value': self.get_date_from_date_time(item.get('LaunchTime'))})
                                    add_tags = self.remove_creation_date(add_tags)
                                    add_tags = self.__add_cluster_id_tag(tags=add_tags)
                                    add_tags = self.__filter_resource_tags_by_add_tags(tags=item.get('Tags'),
                                                                                       search_tags=add_tags)
                                    if add_tags:
                                        cluster_instances.setdefault(cluster_name, []).append(instance_id)
                                        cluster_tags[cluster_name] = add_tags
                                    break
        for cluster_instance_name, instance_ids in cluster_instances.items():
            if self.dry_run == 'no':
                try:
                    self.utils.tag_aws_resources(client_method=self.ec2_client.create_tags, resource_ids=instance_ids,
                                                 tags=cluster_tags.get(cluster_instance_name))
                    logger.info(
                        f'Cluster :: {cluster_instance_name} count: {len(instance_ids)} :: InstanceId :: {instance_ids} :: {cluster_tags.get(cluster_instance_name)}')
                except Exception as err:
                    logger.info(err)
            result_instance_list.extend(instance_ids)
        logger.info(f'cluster_instance :: {len(result_instance_list)} :: {result_instance_list}')
        if not self.cluster_key:
            self.cluster_role(list(cluster_instances.keys()))
            self.cluster_user(list(cluster_instances.keys()))
            self.cluster_s3_bucket(list(cluster_instances.keys()))
        return sorted(result_instance_list)

    def cluster_instance(self):
        """
        This method returns list of cluster's instance according to cluster tag name,
        The instances list is different from other resources
        it will search for full cluster name (including random suffix string) in case of user input cluster name was given
        @return:
        """
        self.cluster_key = self.__init_cluster_name()
        instances_list = self._get_instances_data()
        if instances_list:
            cluster, non_cluster = self.ec2_operations.scan_cluster_or_non_cluster_instance(instances_list)
            if not self.cluster_only:
                ids = self.update_cluster_tags(cluster)
                self.non_cluster_update.non_cluster_update_ec2(non_cluster)
                return ids
            else:
                ids = self.update_cluster_tags(cluster)
                return ids
        else:
            return []

    def cluster_volume(self):
        """
        This method returns list of cluster's volume according to cluster tag name
        @return:
        """
        volumes_data = self.ec2_operations.get_volumes()
        cluster, non_cluster = self.ec2_operations.scan_cluster_non_cluster_resources(volumes_data)
        if not self.cluster_only:
            ids = self.__generate_cluster_resources_list_by_tag(cluster, 'VolumeId')
            self.non_cluster_update.update_volumes(non_cluster)
        else:
            ids = self.__generate_cluster_resources_list_by_tag(cluster, 'VolumeId')
        return ids

    def cluster_ami(self):
        """
        This method returns list of cluster's ami according to cluster tag name
        @return:
        """
        images_data = self.ec2_operations.get_images()
        cluster, non_cluster = self.ec2_operations.scan_cluster_non_cluster_resources(images_data)
        if not self.cluster_only:
            ids = self.__generate_cluster_resources_list_by_tag(cluster, 'ImageId')
            self.non_cluster_update.update_ami(non_cluster)
        else:
            ids = self.__generate_cluster_resources_list_by_tag(cluster, 'ImageId')
        return ids

    def cluster_snapshot(self):
        """
        This method returns list of cluster's snapshot according to cluster tag name
        @return:
        """
        snapshots_data = self.ec2_operations.get_snapshots()
        cluster, non_cluster = self.ec2_operations.scan_cluster_non_cluster_resources(snapshots_data)
        if not self.cluster_only:
            ids = self.__generate_cluster_resources_list_by_tag(cluster, 'SnapshotId')
            self.non_cluster_update.update_snapshots(non_cluster)
        else:
            ids = self.__generate_cluster_resources_list_by_tag(cluster, 'SnapshotId')
        return ids

    def __get_security_group_data(self):
        """
        This method returns security group data
        @return:
        """
        return self.ec2_operations.get_security_groups()

    def cluster_security_group(self):
        """
        This method returns list of cluster's security group according to cluster tag name
        @return:
        """
        security_group_ids = self.__generate_cluster_resources_list_by_tag(
            resources_list=self.__get_security_group_data(),
            input_resource_id='GroupId')
        logger.info(f'cluster_security_group count: {len(sorted(security_group_ids))} {sorted(security_group_ids)}')

    def cluster_elastic_ip(self):
        """
        This method returns list of cluster's elastic ip according to cluster tag name
        @return:
        """
        elastic_ips_data = self.ec2_operations.get_elastic_ips()
        elastic_ips = self.__generate_cluster_resources_list_by_tag(resources_list=elastic_ips_data,
                                                                    input_resource_id='AllocationId')
        logger.info(f'cluster_elastic_ip count: {len(sorted(elastic_ips))} {sorted(elastic_ips)}')
        return sorted(elastic_ips)

    def cluster_network_interface(self):
        """
        This method returns list of cluster's network interface according to cluster tag name
        @return:
        """
        network_interfaces_data = self.ec2_operations.get_network_interface()
        network_interface_ids = self.__generate_cluster_resources_list_by_tag(resources_list=network_interfaces_data,
                                                                              input_resource_id='NetworkInterfaceId',
                                                                              tags='TagSet')
        logger.info(
            f'cluster_network_interface count: {len(sorted(network_interface_ids))} {sorted(network_interface_ids)}')
        return sorted(network_interface_ids)

    def cluster_load_balancer(self):
        """
        This method returns list of cluster's load balancer according to cluster vpc
        @return:
        """
        result_resources_list = []
        load_balancers_data = self.ec2_operations.get_load_balancers()
        for resource in load_balancers_data:
            resource_id = resource['LoadBalancerName']
            tags = self.elb_client.describe_tags(LoadBalancerNames=[resource_id])
            for item in tags['TagDescriptions']:
                if item.get('Tags'):
                    if not self.__validate_existing_tag(item.get('Tags')):
                        for tag in item['Tags']:
                            if any(prefix in tag.get('Key', '') for prefix in self.cluster_prefix):
                                all_tags = []
                                instance_tags = self.__get_cluster_tags_by_instance_cluster(cluster_name=tag.get('Key'))
                                if not instance_tags:
                                    all_tags = self.__append_input_tags(item.get('Tags'))
                                all_tags.extend(instance_tags)
                                all_tags = self.__remove_tags_start_with_aws(all_tags)
                                all_tags = self.__filter_resource_tags_by_add_tags(item.get('Tags'), all_tags)
                                if all_tags:
                                    if self.cluster_name:
                                        if tag['Key'] == self.cluster_key:
                                            try:
                                                if self.dry_run == 'no':
                                                    self.elb_client.add_tags(LoadBalancerNames=[resource_id],
                                                                             Tags=all_tags)
                                                    logger.info(all_tags)
                                            except Exception as err:
                                                logger.exception(f'Tags are already updated, {err}')
                                            result_resources_list.append(resource_id)
                                        break
                                    else:
                                        if self.dry_run == 'no':
                                            try:
                                                self.elb_client.add_tags(LoadBalancerNames=[resource_id], Tags=all_tags)
                                                logger.info(all_tags)
                                            except Exception as err:
                                                logger.exception(f'Tags are already updated, {err}')
                                        result_resources_list.append(resource_id)
                                        break
                                break
        logger.info(
            f'cluster_load_balancer count: {len(sorted(result_resources_list))} {sorted(result_resources_list)}')
        return sorted(result_resources_list)

    def cluster_load_balancer_v2(self):
        """
        This method returns list of cluster's load balancer according to cluster vpc
        @return:
        """
        result_resources_list = []
        load_balancers_data = self.ec2_operations.get_load_balancers_v2()
        for resource in load_balancers_data:
            resource_id = resource['LoadBalancerArn']
            tags = self.elbv2_client.describe_tags(ResourceArns=[resource_id])
            for item in tags['TagDescriptions']:
                if item.get('Tags'):
                    if not self.__validate_existing_tag(item.get('Tags')):
                        for tag in item['Tags']:
                            if any(prefix in tag.get('Key', '') for prefix in self.cluster_prefix):
                                all_tags = []
                                instance_tags = self.__get_cluster_tags_by_instance_cluster(cluster_name=tag.get('Key'))
                                if not instance_tags:
                                    all_tags = self.__append_input_tags(item.get('Tags'))
                                all_tags.extend(instance_tags)
                                all_tags = self.__remove_tags_start_with_aws(all_tags)
                                all_tags = self.__filter_resource_tags_by_add_tags(item.get('Tags'), all_tags)
                                if all_tags:
                                    if self.cluster_name:
                                        if tag['Key'] == self.cluster_key:
                                            try:
                                                if self.dry_run == 'no':
                                                    self.elbv2_client.add_tags(ResourceArns=[resource_id],
                                                                               Tags=all_tags)
                                                    logger.info(all_tags)
                                            except Exception as err:
                                                logger.exception(f'Tags are already updated, {err}')
                                            result_resources_list.append(resource_id)
                                        break
                                    else:
                                        if self.dry_run == 'no':
                                            try:
                                                self.elbv2_client.add_tags(ResourceArns=[resource_id], Tags=all_tags)
                                                logger.info(all_tags)
                                            except Exception as err:
                                                logger.exception(f'Tags are already updated, {err}')
                                        result_resources_list.append(resource_id)
                                        break
                                break
        logger.info(
            f'cluster_load_balancer_v2 count: {len(sorted(result_resources_list))} {sorted(result_resources_list)}')
        return sorted(result_resources_list)

    def cluster_vpc(self):
        """
        This method returns list of cluster's vpc according to cluster tag name
        @return:
        """
        vpcs_data = self.ec2_operations.get_vpcs()
        vpc_ids = self.__generate_cluster_resources_list_by_tag(resources_list=vpcs_data, input_resource_id='VpcId')
        logger.info(f'cluster_vpc count: {len(sorted(vpc_ids))} {sorted(vpc_ids)}')
        self.cluster_network_acl()
        return sorted(vpc_ids)

    def get_cluster_vpc(self):
        """
        This method get cluster vpc ids and it's tags.
        Missing OpenShift Tags for it based on VPCs
        @return:
        """
        vpcs_data = self.ec2_operations.get_vpcs()
        vpc_ids = {}
        for vpc in vpcs_data:
            if vpc.get('Tags'):
                for tag in vpc.get('Tags'):
                    if any(prefix in tag.get('Key', '') for prefix in self.cluster_prefix):
                        vpc_ids[vpc.get('VpcId')] = [tag for tag in vpc.get('Tags') if tag.get('Key') != 'Name']
                        break
        return vpc_ids

    def cluster_subnet(self):
        """
        This method returns list of cluster's subnet according to cluster tag name
        @return:
        """
        subnets_data = self.ec2_operations.get_subnets()
        subnet_ids = self.__generate_cluster_resources_list_by_tag(resources_list=subnets_data,
                                                                   input_resource_id='SubnetId')
        logger.info(f'cluster_subnet count: {len(sorted(subnet_ids))} {sorted(subnet_ids)}')
        return sorted(subnet_ids)

    def cluster_route_table(self):
        """
        This method returns list of cluster's route table according to cluster tag name
        @return:
        """
        route_tables_data = self.ec2_operations.get_route_tables()
        route_table_ids = self.__generate_cluster_resources_list_by_tag(resources_list=route_tables_data,
                                                                        input_resource_id='RouteTableId')
        logger.info(f'cluster_route_table count: {len(sorted(route_table_ids))} {sorted(route_table_ids)}')
        return sorted(route_table_ids)

    def cluster_internet_gateway(self):
        """
        This method returns list of cluster's route table internet gateway according to cluster tag name
        @return:
        """
        internet_gateways_data = self.ec2_operations.get_internet_gateways()
        internet_gateway_ids = self.__generate_cluster_resources_list_by_tag(resources_list=internet_gateways_data,
                                                                             input_resource_id='InternetGatewayId')
        logger.info(
            f'cluster_internet_gateway count: {len(sorted(internet_gateway_ids))} {sorted(internet_gateway_ids)}')
        return sorted(internet_gateway_ids)

    def cluster_dhcp_option(self):
        """
        This method returns list of cluster's dhcp option according to cluster tag name
        @return:
        """
        dhcp_options_data = self.ec2_operations.get_dhcp_options()
        dhcp_ids = self.__generate_cluster_resources_list_by_tag(resources_list=dhcp_options_data,
                                                                 input_resource_id='DhcpOptionsId')
        logger.info(f'cluster_dhcp_option count: {len(sorted(dhcp_ids))} {sorted(dhcp_ids)}')
        return sorted(dhcp_ids)

    def cluster_vpc_endpoint(self):
        """
        This method returns list of cluster's vpc endpoint according to cluster tag name
        @return:
        """
        vpc_endpoints_data = self.ec2_operations.get_vpce()
        vpc_endpoint_ids = self.__generate_cluster_resources_list_by_tag(resources_list=vpc_endpoints_data,
                                                                         input_resource_id='VpcEndpointId')
        logger.info(f'cluster_vpc_endpoint count: {len(sorted(vpc_endpoint_ids))} {sorted(vpc_endpoint_ids)}')
        return sorted(vpc_endpoint_ids)

    def cluster_nat_gateway(self):
        """
        This method returns list of cluster's nat gateway according to cluster tag name
        @return:
        """
        nat_gateways_data = self.ec2_operations.get_nat_gateways()
        nat_gateway_id = self.__generate_cluster_resources_list_by_tag(resources_list=nat_gateways_data,
                                                                       input_resource_id='NatGatewayId')
        logger.info(f'cluster_nat_gateway count: {len(sorted(nat_gateway_id))} {sorted(nat_gateway_id)}')
        return sorted(nat_gateway_id)

    def cluster_network_acl(self):
        """
        This method returns list of cluster's network acl according to cluster vpc id
        Missing OpenShift Tags for it based on VPCs
        @return:
        """
        network_acls_data = self.ec2_operations.get_nacls()
        network_acl_ids = self.__generate_cluster_resources_list_by_vpc(resources_list=network_acls_data,
                                                                        input_resource_id='NetworkAclId')
        logger.info(f'cluster_network_acl count: {len(network_acl_ids)}, {network_acl_ids}')
        return sorted(network_acl_ids)

    def cluster_role(self, cluster_names: list = []):
        """
        This method returns list of cluster's role according to cluster name
        @param cluster_names:
        @return:
        """
        # tag_role
        result_role_list = []
        # if cluster_key exit
        if self.cluster_name:
            cluster_names.append(self.cluster_name)
        if cluster_names:
            for cluster_name in cluster_names:
                cluster_key = self.cluster_name if self.cluster_key else cluster_name
                if cluster_key:
                    # starts with cluster name, search for specific role name for fast scan (a lot of roles)
                    role_name_list = []
                    roles = self.iam_operations.get_roles()
                    for role in roles:
                        if cluster_key in role.get('RoleName'):
                            role_name_list.append(role.get('RoleName'))
                    if role_name_list:
                        for role_name in role_name_list:
                            try:
                                role = self.iam_client.get_role(RoleName=role_name)
                                role_data = role['Role']
                                all_tags = []
                                instance_tags = self.__get_cluster_tags_by_instance_cluster(
                                    cluster_name=cluster_key)
                                if not instance_tags:
                                    all_tags = self.__append_input_tags(role_data.get('Tags'))
                                else:
                                    all_tags.extend(instance_tags)
                                all_tags = self.__remove_tags_start_with_aws(all_tags)
                                all_tags = self.__filter_resource_tags_by_add_tags(role_data.get('Tags'), all_tags)
                                if all_tags:
                                    if self.dry_run == 'no':
                                        try:
                                            self.iam_client.tag_role(RoleName=role_name, Tags=all_tags)
                                            logger.info(all_tags)
                                        except Exception as err:
                                            logger.exception(f'Tags are already updated, {err}')
                                    result_role_list.append(role_data['Arn'])
                            except Exception as err:
                                logger.exception(f'Missing cluster role name: {role_name}, {err}')
                    else:
                        logger.info(f'Missing role for cluster {cluster_name}')
        logger.info(f'cluster_role count: {len(sorted(result_role_list))} {sorted(result_role_list)}')
        return sorted(result_role_list)

    def cluster_user(self, cluster_names: list = []):
        """
        This method returns list of cluster's user according to cluster name
        @param cluster_names:
        @return:
        """
        # tag_user
        result_user_list = []
        if self.cluster_name:
            cluster_names.append(self.cluster_name)
        if cluster_names:
            for cluster_name in cluster_names:
                users = self.iam_operations.get_users()
                # return self.__generate_cluster_resources_list_by_tag(resources_list=users_data,
                #                                                      input_resource_id='UserId')
                cluster_name = self.cluster_name if self.cluster_key else cluster_name
                for user in users:
                    user_name = user['UserName']
                    if cluster_name in user_name:
                        user_data = self.iam_client.get_user(UserName=user_name)
                        data = user_data['User']
                        if data.get('Tags'):
                            # search that not exist permanent tags in the resource
                            if not self.__validate_existing_tag(data.get('Tags')):
                                for tag in data['Tags']:
                                    if cluster_name in tag['Key']:
                                        all_tags = []
                                        instance_tags = self.__get_cluster_tags_by_instance_cluster(
                                            cluster_name=cluster_name)
                                        if not instance_tags:
                                            all_tags = self.__append_input_tags(data.get('Tags'))
                                        all_tags.extend(instance_tags)
                                        all_tags = self.__remove_tags_start_with_aws(all_tags)
                                        all_tags = self.__filter_resource_tags_by_add_tags(data.get('Tags'), all_tags)
                                        if all_tags:
                                            if self.dry_run == 'no':
                                                try:
                                                    self.iam_client.tag_user(UserName=user_name, Tags=all_tags)
                                                    logger.info(all_tags)
                                                except Exception as err:
                                                    logger.info(err)
                                            result_user_list.append(user_name)
                                        break
        logger.info(f'cluster_user count: {len(sorted(result_user_list))} {sorted(result_user_list)}')
        return sorted(result_user_list)

    def __filter_resource_tags_by_add_tags(self, tags: list, search_tags: list):
        """
        This method filters the tags by cluster and adding tags.
        @param tags:
        @param search_tags:
        @return:
        """
        add_tags = []
        if tags:
            for search_tag in search_tags:
                found = False
                for tag in tags:
                    if tag.get('Key') == search_tag.get('Key') and tag.get('Value') != 'NA':
                        found = True
                        break
                if not found:
                    add_tags.append(search_tag)
        else:
            add_tags.extend(search_tags)
        filter_tags = {}
        for tag in add_tags:
            key = tag.get('Key')
            value = tag.get('Value')
            if key in filter_tags and filter_tags[key].get('Value') == self.NA_VALUE:
                filter_tags[key] = {'Key': key, 'Value': value}
            else:
                filter_tags[key] = {'Key': key, 'Value': value}
        return list(filter_tags.values())

    def __remove_launchTime(self, tags: list):
        """
        This method removes the launch time form the instance tags
        @param tags:
        @return:
        """
        return [tag for tag in tags if tag.get('Key') != 'LaunchTime']

    def cluster_s3_bucket(self, cluster_names: list = []):
        """
        This method returns list of cluster's s3 bucket according to cluster name
        @param cluster_names:
        @return:
        """
        bucket_result_list = []
        response = self.s3_client.list_buckets()
        # if cluster_key exit
        if self.cluster_name:
            cluster_names.append(self.cluster_name)
        if cluster_names:
            for cluster_name in cluster_names:
                cluster_key = self.cluster_name if self.cluster_key else cluster_name
                if cluster_key:
                    for bucket in response['Buckets']:
                        # starts with cluster name
                        if bucket['Name'].startswith(cluster_key):
                            try:
                                bucket_tags = self.s3_client.get_bucket_tagging(Bucket=bucket.get('Name'))
                            except Exception as err:
                                logger.info(err)
                                bucket_tags = {}
                            if bucket_tags:
                                bucket_tags = bucket_tags.get('TagSet')
                                # search that not exist permanent tags in the resource
                                if not self.__validate_existing_tag(bucket_tags):
                                    add_tags = []
                                    instance_tags = self.__get_cluster_tags_by_instance_cluster(
                                        cluster_name=cluster_name)
                                    if not instance_tags:
                                        add_tags = self.__append_input_tags(bucket_tags)
                                    add_tags.extend(instance_tags)
                                    add_tags = self.__remove_tags_start_with_aws(add_tags)
                                    add_tags = self.__filter_resource_tags_by_add_tags(bucket_tags, add_tags)
                                    if add_tags:
                                        if self.dry_run == 'no':
                                            try:
                                                add_tags.extend(bucket_tags)
                                                self.s3_client.put_bucket_tagging(Bucket=bucket.get('Name'),
                                                                                  Tagging={'TagSet': add_tags})
                                                logger.info(add_tags)
                                            except Exception as err:
                                                logger.info(err)
                                        bucket_result_list.append(bucket['Name'])
        logger.info(f'cluster_s3_bucket count: {len(sorted(bucket_result_list))} {sorted(bucket_result_list)}')
        return sorted(bucket_result_list)



import importlib
import inspect
from datetime import datetime

from cloud_governance.common.logger.init_logger import logger
from cloud_governance.policy.policy_operations.aws.zombie_non_cluster.run_zombie_non_cluster_policies import NonClusterZombiePolicy


class ZombieNonClusterPolicies(NonClusterZombiePolicy):

    def __init__(self):
        super().__init__()

    def run(self):
        """
        This method run the zombie non-cluster policies class
        @return:
        """
        logger.info(f'account={self._account}, region={self._region}, policy={self._policy}, dry_run={self._dry_run}')
        zombie_non_cluster_policy_module = importlib.import_module(f'cloud_governance.policy.aws.{self._policy}')

        for cls in inspect.getmembers(zombie_non_cluster_policy_module, inspect.isclass):
            if self._policy.replace('_', '') == cls[0].lower():
                response = cls[1]().run()
                if isinstance(response, str):
                    logger.info(f'key: {cls[0]}, Response: {response}')
                else:
                    logger.info(f'key: {cls[0]}, count: {len(response)}, {response}')
                    policy_result = response

                    if self._es_operations.check_elastic_search_connection():
                        if policy_result:
                            if len(policy_result) > 500:
                                self._es_operations.upload_data_in_bulk(data_items=policy_result.copy(),
                                                                        index=self._es_index)
                            else:
                                for policy_dict in policy_result:
                                    policy_dict['region_name'] = self._region
                                    policy_dict['account'] = self._account
                                    self._es_operations.upload_to_elasticsearch(data=policy_dict.copy(), index=self._es_index)
                            logger.info(f'Uploaded the policy results to elasticsearch index: {self._es_index}')
                        else:
                            logger.info(f'No data to upload on @{self._account}  at {datetime.utcnow()}')
                    else:
                        logger.error('ElasticSearch host is not pingable, Please check ')

                    if self._policy_output:
                        # if self._policy not in ('ec2_idle', 'ebs_in_use', 'ec2_run', 's3_inactive', 'unused_access_key', 'zombie_snapshots', 'nat_gateway_unused'):
                        #     beautify_data = self._beautify_upload_data(upload_resource_data=response)
                        #     policy_result = {'count': len(beautify_data), self._policy: beautify_data}
                        logger.info(policy_result)
                        self._s3operations.save_results_to_s3(policy=self._policy.replace('_', '-'), policy_output=self._policy_output, policy_result=policy_result)
